#include <QEventLoop>
#include <QScriptEngine>
#include <QByteArray>
#include <QSettings>
#include <QStringList>
#include <QApplication>
#include <QUrl>
#include <QFile>
#include <QDebug>
#include <QMessageBox>
#include <QDesktopWidget>
#include <QRegExp>

#include "logger.h"
#include "ui/widgets/browser/browser.h"
#include "ui/widgets/document.h"
#include "ui/widgets/gif.h"
#include "ui/widgets/idle.h"
#include "ui/widgets/image.h"
#include "ui/widgets/pdf.h"
#include "ui/widgets/video.h"
#include "ui/widgets/webpage.h"
#include "ui/widgets/externalbrowser.h"
#include "playlistparser.h"
#include "downloader.h"
#include "dispatcher.h"
#include "maintenance.h"

#include "ui/exitpassworddialog.h"
#include "utils.h"

Dispatcher::Dispatcher(Viewer *viewer, const CommandLineParser *cmdparse)
    : QObject()
    , settings(NULL)
    , cleanser(NULL)
    , m_status(Dispatcher::StatusNone)
    , m_currentRunnable(NULL)
    , m_currentWidget(NULL)
    , m_viewer(viewer)
    , remote(NULL)
    , m_plRefresh(false)
    , m_updateMode(true)
    , m_pl(NULL)
    , m_ppl(NULL)
    , m_loop(NULL)
    , m_loopIdx(0)
    , m_listIdx(0)
    , m_idling(false)
    , m_is_valid(false)
    , m_oldPl(true)
{
    settings = Settings::load(false, cmdparse);

#if 0
    if (settings->versionChanged())
    {
        QMessageBox::warning(
            viewer,
            tr("Software version changed"),
            tr("You changed software version. Please reset your linutop settings."));
        m_is_valid = false;
    }
    else {
#endif
        m_is_valid = true;
        start_sec = QDateTime::currentDateTime();
#if 0
    }
#endif

    remote = new Remote();
    cleanser = new Cleanser(settings);
    QObject::connect(remote, SIGNAL(restart()),            this,     SLOT(startSlideshow()));
    QObject::connect(remote, SIGNAL(next()),               this,     SLOT(next()));
    QObject::connect(remote, SIGNAL(goTo(int,int)),        this,     SLOT(goTo(int,int)));
    QObject::connect(remote, SIGNAL(goTo(QString)),        this,     SLOT(goTo(QString)));
    QObject::connect(remote, SIGNAL(pause()),              m_viewer, SLOT(pause()));
    QObject::connect(remote, SIGNAL(screenshot()),         this,     SLOT(shootScreen()));
    QObject::connect(remote, SIGNAL(text(QString,int)),    m_viewer, SLOT(showText(QString,int)));
    QObject::connect(remote, SIGNAL(readUrl(QString,int)), m_viewer, SLOT(showUrl(QString,int)));

    m_mode = settings->get("main/mode").toString();

    QObject::connect(viewer, SIGNAL(restart()),           this, SLOT(startSlideshow()));
    QObject::connect(viewer, SIGNAL(end()),               this, SLOT(next()));
    QObject::connect(viewer, SIGNAL(mouseEvent()),        this, SLOT(onMouseEvent()));
    QObject::connect(viewer, SIGNAL(leftclick()),         this, SLOT(onLeftClick()));
    QObject::connect(viewer, SIGNAL(terminate()),         this, SLOT(finalize()));
    QObject::connect(viewer, SIGNAL(startInlineRemote()), this, SLOT(startInlineRemote()));
    QObject::connect(viewer, SIGNAL(readyToShoot()),      this, SLOT(shootScreen()), Qt::QueuedConnection);
#ifdef __arm__
    endIdle();
    connect(viewer, SIGNAL(terminate()),         this, SLOT(endIdle()));
#endif

    if(settings->get("main/fullscreen").toBool())
    {
        QRect rec = QApplication::desktop()->screenGeometry();
        int height = rec.height();
        int width = rec.width();
        m_viewer->setFixedSize(width,height);
    }

    // Disable screensaver when option is selected.
    if (settings->get("main/no_screensaver").toBool())
    {
        QProcess p;
//        QStringList args;
//        args << "xscreensaver-command";
//        QObject::connect(
//                    &p,   SIGNAL(finished(int, QProcess::ExitStatus)),
//                    this, SLOT(startNoScreensaver(int, QProcess::ExitStatus)));
//        p.start("which", args);
//        p.waitForFinished();

        //Disable xscreensaver deamon and suppress ctrl alt suppr shortcut
        p.start("pkill -9 xscreensaver");
        p.waitForFinished();
    }

    // Delete previous screenshot if needed.
    {
        if (QFile::exists("/tmp/screen.png"))
            QFile::remove("/tmp/screen.png");
    }
}

Dispatcher::~Dispatcher(void)
{
    // Kill the created server.
    if (!remotePid.isEmpty())
    {
        QProcess cmdKill;
        cmdKill.start("sudo", QStringList() << "kill" << "-9" << remotePid.toStdString().c_str());
        cmdKill.waitForFinished();
        qDebug() << "Remote killed:" << remotePid.toStdString().c_str();
    }
    else
    {
        qDebug() << "No remote to kill.";
    }
}

void Dispatcher::startNoScreensaver(
        int exitCode, QProcess::ExitStatus /*exitStatus*/)
{ if (exitCode == 0) disableScreensaver(); }

void Dispatcher::onScreensaverTimeout()
{
    QProcess p;
    QStringList args;
    args << "-deactivate";
    p.start("xscreensaver-command", args);
    p.waitForFinished();
}

void Dispatcher::disableScreensaver()
{
    connect(&m_ssTimer, SIGNAL(timeout()), this, SLOT(onScreensaverTimeout()));
    m_ssTimer.start(30000);
}


void Dispatcher::onPlTimeout()
{
    m_plRefresh = true;
    QString updateMode = settings->get(
                "slideshow/playlist_update_mode").toString();
    if (updateMode == "serverLinutopTV" || updateMode == "serverLinutopTVPrivate")
    {
        QDateTime lastUpdate = getLastEdit();
        QFile file(QDir::homePath() + "/.linutop-kiosk/playlist_last_timestamp.txt");
        if(file.open(QIODevice::ReadOnly))
        {
            QString lastEdit = file.readAll();
            if(lastEdit== lastUpdate.toString() || lastEdit.isNull())
                m_plRefresh = false;
        }
        file.close();
    }
}

bool Dispatcher::startup(void)
{
    QString updateMode = settings->get(
                "slideshow/playlist_update_mode").toString();

    if (updateMode == "serverLinutopTV" || updateMode == "serverLinutopTVPrivate")
    {
        connect(&m_plTimer, SIGNAL(timeout()), this, SLOT(onPlTimeout()));
        int timeout = settings->get("slideshow/refresh_after_minutes").toInt();
        m_plTimer.start(60000 * timeout);
    }

    if (m_mode == "kiosk" && settings->get("main/start_with_session").toBool())
        return startBrowser();
    else
        startSlideshow();

    return true;
}

void Dispatcher::onMouseEvent()
{
    if (m_mode != "kiosk") {
        return;
    }

    m_currentRunnable->stop();
    m_loopIdx = -1;
    m_loopIdxList.replace(m_listIdx, m_loopIdx);
    startBrowser();
}

void Dispatcher::onLeftClick()
{
    next();
}

void Dispatcher::onReconnectMouseEventTimeout()
{
    connect(m_viewer, SIGNAL(mouseEvent()), this, SLOT(onMouseEvent()));
}

void Dispatcher::reconnectMouseEvent()
{
    QTimer::singleShot(1000, this, SLOT(onReconnectMouseEventTimeout()));
}

bool Dispatcher::next(void)
{
    if (m_status == Dispatcher::StatusNone) return startup();

    else if (m_status == Dispatcher::StatusBrowser)
    {
        disconnect(m_viewer, SIGNAL(mouseEvent()), this, SLOT(onMouseEvent()));
        execHook("session-stop");
        afterBrowser();
    }

    else if (m_status == Dispatcher::StatusAfterBrowser)
    {
        if (settings->get("main/after_web_session").toString() == "session")
        {
            Logger::logInfo("STARTING: web session");
            startBrowser();
        }
        else
        {
            reconnectMouseEvent();
            startSlideshow();
        }
    }
    else showLoopNext();

    return true;
}

void Dispatcher::shootScreen(void)
{
    if (!m_viewer->currentWidget()) return;
    QScreen *screen = QGuiApplication::primaryScreen();
    if (const QWindow *window = m_viewer->windowHandle())
        screen = window->screen();
    if (!screen) return;
    screen->grabWindow(m_viewer->winId()).save("/tmp/screen.png", "PNG");
    //m_viewer->grab().save("/tmp/screen.png", "PNG");
}

bool Dispatcher::view_next()
{
    return m_viewer->next(m_currentWidget);
}

void Dispatcher::afterBrowser()
{
    m_status = Dispatcher::StatusAfterBrowser;
    QString uri = settings->get("main/post_session_uri").toString();
    if (uri.isEmpty())
    {
        next();
        return;
    }
    int time = settings->get("main/post_session_uri_time").toInt();
    PlaylistItem item(uri, PlaylistItem::TypeWebpage, time);
    Webpage* page = new Webpage(&item, m_viewer);
    m_currentWidget = page;
    m_currentRunnable = page;
    view_next();
}

bool Dispatcher::startBrowser(QUrl url)
{
    if (settings->get("main/external_browser").toBool())
    {
        bool use_password = settings->get("main/admin_passwd").toBool();

        QScopedPointer<panel::Manager> panelManager(
                    panel::Manager::create(true, use_password));
        QScopedPointer<SystemShortcutManager> shortcutManager(
                    new SystemShortcutManager(true, use_password));

        QProcess process;
        process.start(settings->get("main/external_browser_cmd").toString() +
                      " " + settings->get("main/default_url").toString());
        m_status = Dispatcher::StatusBrowser;
        process.waitForFinished(-1);

        if (use_password)
        {
            ExitPasswordDialog d;
            QString password = d.run();
            if (password.isEmpty() || !Utils::validateAdminPassword(password))
                startBrowser();
        }

        return false;
    }
    else
    {
        execHook("session-start");
        Logger::logInfo("STARTING: web session");
        m_status = Dispatcher::StatusBrowser;
        Browser* browser = new Browser;
        m_currentWidget = browser;
        m_currentRunnable = browser;
        if (!url.isEmpty()) browser->setHome(url.toString());
        view_next();
    }

    return true;
}

void Dispatcher::execHook(QString hook)
{
    QString hookFile = settings->hookDirname() + QString("/%1").arg(hook);
    QProcess p;
    p.start(hookFile);
    p.waitForFinished();
}

QDateTime Dispatcher::getLastEdit()
{
    QUrl url = QUrl(settings->get("slideshow/pages").toString());
    qDebug() << "Loading pages from:" << url.toString().toStdString().c_str();
    QRegExp Exp(".*/p(/[\\w-]+)/.*");

    Exp.setMinimal(true);
    int pos = Exp.indexIn(url.toString());
    if(pos > -1){

    }
    QStringList list = Exp.cap(1).split("-");
    QString login = list.at(0);
    QString screen = "/" + list.at(1);
    QUrl urlReq;
    urlReq.setScheme(url.scheme());
    urlReq.setHost(url.host());
    QString function = "";
    if(m_lastUpdate.toString() == "")
        function = "lastEdit";
    else
        function = "getEdit";

    urlReq.setPath("/webapp/db/"+function+"/playlist/"+((XmlPlaylistParser *)m_pl)->getName()+login+screen);

    qDebug() << "Request URL: " << urlReq.toString().toStdString().c_str();

    QNetworkRequest req(urlReq);
    QNetworkAccessManager nam;
    QNetworkReply * reply = nam.get(req);

    QEventLoop loop;
    connect(reply, SIGNAL(finished()), &loop, SLOT(quit()));
    loop.exec();
    loop.deleteLater();

    qDebug() << "Reply finished";

    QScriptEngine engine;
    QString res = reply->readAll();
    QScriptValue v = engine.evaluate(res);
    res = "";
    Q_FOREACH (QString number, v.toVariant().toStringList())
        res += number+";";

    return QDateTime::fromString(res, "yyyy;MM;dd;hh;mm;ss;");
}

void Dispatcher::startSlideshow()
{
    cleanser->check();
    if (settings->get("remote/enabled").toBool()) startRemote();

    QUrl url;

    settings = Settings::load(true, settings->getCommandLineParser());

    QString updateMode = settings->get(
                "slideshow/playlist_update_mode").toString();

    if (updateMode != "localPlaylist")
        url = QUrl(settings->get("slideshow/pages").toString());
    else
        url = QUrl(QString("file://%1").arg(settings->playlistFilename()));

    bool ok = false;
    int default_time = settings->get("slideshow/default_time").toInt(&ok);
    m_pl = PlaylistParser::get((ok ? default_time : 5), url, m_viewer);

    if (updateMode == "serverLinutopTV" || updateMode == "serverLinutopTVPrivate")
    {
        m_lastUpdate = getLastEdit();
        QFile file(QDir::homePath() + "/.linutop-kiosk/playlist_last_timestamp.txt");
        QByteArray fileData;
        if (file.open(QIODevice::ReadWrite))
        {
            fileData = file.readAll();
            QString data(fileData);
            if(data != m_lastUpdate.toString().toUtf8())
            {
                file.resize(0);
                file.write(m_lastUpdate.toString().toUtf8());
                m_updateMode = true;
            }
            else
                m_updateMode = false;
        }
        file.close();
    }

    Q_FOREACH (PlaylistLoop* loop, m_pl->getLoops())
    {
        if (loop->downloadable())
        {
            Downloader *d = new Downloader(loop, m_updateMode);
            QEventLoop el;
            connect(d,SIGNAL(downloadDone()),&el,SLOT(quit()));
            connect(d,SIGNAL(downloadNotDone()),this,SLOT(startSlideshow()));
            connect(d,SIGNAL(downloadNotDone()),&el,SLOT(quit()));
            d->start();
            if(!d->isDone())el.exec();
            d->deleteLater();
        }
    }
    if(m_updateMode)qDebug()<<"Loop downloaded";
    else qDebug() << "Loop not downloaded";

    //We fill the list which contains the index of current item in each loops
    if(m_loopIdxList.count() == 0 || updateMode == "localPlaylist" || m_updateMode)
    {
        if(m_updateMode)
            m_loopIdxList.clear();
        if(m_loopIdxList.isEmpty())
        {
            qDebug() << "NB LOOPS : " << m_pl->getLoops().count();
            for(int i=0; i < m_pl->getLoops().last()->id().toInt()+1; i++)
                m_loopIdxList.append(-1);
        }
    }
    m_loopIdx = -1;
    showLoopNext();
}

void Dispatcher::goTo(int index, int loop)
{
    m_loop = m_pl->getLoops().at(loop);
    m_listIdx = m_loop->id().toInt();
    m_loopIdx = index - 1;
    m_loopIdxList.replace(m_listIdx, m_loopIdx);
    showLoopNext();
}

void Dispatcher::goTo(QString str)
{
    // We suspend the actual playlist and create a RAM one to display directly an item.
    QString content = QString("#Linutop playlist\nfile:/" + str + " 1\n");
    m_loop = new PlaylistLoop(1, content, QString(""));
    m_loopIdx = -1;
    m_loopIdxList.replace(m_listIdx, m_loopIdx);
    this->showLoopNext();
    m_viewer->pause();
}

void Dispatcher::showLoopNext()
{
    QString updateMode = settings->get(
                "slideshow/playlist_update_mode").toString();
    m_status = Dispatcher::StatusPlaylist;
    m_loop = m_pl->selectLoop(start_sec);

    if(updateMode != "localPlaylist")
        m_listIdx = m_loop->id().toInt();
    else
        m_listIdx = 0;



    m_loopIdx = m_loopIdxList.at(m_listIdx);
    m_loopIdxList.replace(m_listIdx, ++m_loopIdx);

    if (m_plRefresh)
    {
        m_plRefresh = false;
        startSlideshow();
        return;
    }

    if (m_loopIdx >= m_loop->count())
    {
        /* We need to download the playlist to check what's changed */
        m_pl->refresh();
        m_loop = m_pl->selectLoop(start_sec);
        m_listIdx = m_loop->id().toInt();

        if (m_loop->count() == 0)
        {
            qDebug() << "Replace maintenance file";
            m_loop->items().append(
                        new PlaylistItem(
                            "file://:/image/nothing_to_display.png",
                            PlaylistItem::TypeUndefined, 5, false));
        }
        m_loopIdx = 0; // no download, no event, so continue
        m_loopIdxList.replace(m_listIdx, m_loopIdx);
    }

    PlaylistItem *item;
    if (m_loop->count() > m_loopIdx)
    {
        if(m_loop->at(m_loopIdx)->type() == PlaylistItem::TypeMaintenance)
        {
            qDebug() << "Start maintenance.";
            Maintenance m;
            m.startMaintenance(m_loop->at(m_loopIdx)->uri());
            m_loop->items().append(
                        new PlaylistItem(
                            "file://:/image/nothing_to_display.png",
                            PlaylistItem::TypeUndefined, 60, false));
        }

        item = m_loop->at(m_loopIdx);
        start_sec = QDateTime::currentDateTime();

        if((updateMode == "localPlaylist")
                && item->uri().contains("Screen OFF")) //Mise en place idle sur la boucle locale
        {
            QDateTime start_idle = m_loop->getFromIdle(item->uri());
            QDateTime stop_idle = m_loop->getToIdle(item->uri());

            if(start_idle.time() > stop_idle.time()) {
                if(start_idle.time() <= start_sec.time() && start_sec.time() < QTime(23,59,59)) {
                    qDebug() << "Wake up : tomorrow";
                    stop_idle = stop_idle.addDays(1);
                }
                else if (QTime(0,0,0) <= start_sec.time() && start_sec.time() < stop_idle.time()) {
                    qDebug() << "Wake up : today";
                    start_idle =  start_idle.addDays(-1);
                }
            }

            if (start_idle > start_sec || start_sec > stop_idle )
            {
                m_loopIdx++;
                if (m_loopIdx >= m_loop->count())
                {
                    m_loopIdx = 0;
                }
                m_loopIdxList.replace(m_listIdx, m_loopIdx);
                item = m_loop->at(m_loopIdx);
            }
            else {
                item->setTime(start_sec.secsTo(stop_idle));
            }

        }
    }

    else
        return;

    qDebug() << item->uri() << item->type() << item->href();

    QStringList listExclusions;
   listExclusions << ".docx" << ".ppt" << ".doc" <<".pptx" << ".psd" << ".db" << ".ods" << ".tmp" <<".swf" <<".cfg"<<".sh";

    for (int i=0;i<listExclusions.size();i++) {


    if (item->uri().contains(listExclusions[i]))
    {
         showLoopNext();
   return;
    }
}
    if (item->href())
    {
        showLoopNext();
        return;
    }

    Runnable *prevRunnable = m_currentRunnable;

    switch (item->type())
    {
        case PlaylistItem::TypeIdle: {
            Idle* idle = new Idle(item, m_viewer);
            m_currentWidget = idle;
            m_currentRunnable = idle;
            break;
        }
        case PlaylistItem::TypeImage: {
            Image* image = new Image(item, m_viewer);
            m_currentWidget = image;
            m_currentRunnable = image;
            break;
        }
        case PlaylistItem::TypeGif: {
            Gif* gif = new Gif(item, m_viewer);
            m_currentWidget = gif;
            m_currentRunnable = gif;
            break;
        }
        case PlaylistItem::TypeVideo: {
            Video* video = new Video(item, m_viewer);
            m_currentWidget = video;
            m_currentRunnable = video;
            break;
        }
        case PlaylistItem::TypePdf: {
            PDF* pdf = new PDF(item, m_viewer);
            m_currentWidget = pdf;
            m_currentRunnable = pdf;
            break;
        }
        case PlaylistItem::TypeWebpage: {
            if (settings->get("slideshow/external_browser").toBool()) {
                ExternalBrowser* browser = new ExternalBrowser(item, m_viewer);
                m_currentWidget = browser;
                m_currentRunnable = browser;
            }
            else {
                Webpage* webpage = new Webpage(item, m_viewer);
                m_currentWidget = webpage;
                m_currentRunnable = webpage;
            }
            break;
        }
        case PlaylistItem::TypeDocument: {
            Document* doc = new Document(item, m_viewer);
            m_currentWidget = doc;
            m_currentRunnable = doc;
            break;
        }
        default: {
            next();
            return;
        }
    }

    if ((dynamic_cast<Idle *>(prevRunnable) != NULL)
        && (item->type() != PlaylistItem::TypeIdle))
    {
        qDebug() << "Stopping idle";
        QProcess p;
#ifdef __arm__
        endIdle();
#else
        p.start("xset", QStringList() << "dpms" << "force" << "on");
        p.waitForFinished();
#endif
    }

    if (!view_next())
    {
        --m_loopIdx;
        showLoopNext();
    }
}

void Dispatcher::startRemote()
{
    // Kill remaining surviving servers.
    {
        QProcess cmdKill;
        cmdKill.start("sudo", QStringList() << "pkill" << "-9" << "linutop-remote");
        cmdKill.waitForFinished();
    }

    // Start the new server.
    {
        QStringList args;
#ifdef DEBUG
        args << LINUTOP_REMOTE_CMD << "--debug";
#else
        args << "linutop-remote";
#endif // DEBUG
        args << getenv("HOME");
        args << getenv("USER");
        qDebug() << "Start remote: sudo"
                 << args.join(" ").toStdString().c_str();
        QProcess server;
        //server.setWorkingDirectory("/");
        server.start("sudo", args);
        server.waitForFinished();

        QProcess pid_serv;
        pid_serv.start("sh", QStringList()<<"-c" << "ps -aux | grep server.py | grep python | sed -nr 1p | awk '{print $2}'" );
        pid_serv.waitForFinished();

         QByteArray output = pid_serv.readAllStandardOutput();

   qDebug()<<"TAILLE DU PID" << output.length();
   int taille = output.length()-1;

        qDebug() << "OUTPUT:"<<output.left(taille);
        remotePid = QString(output.left(taille));


        qDebug() << "Remote started, pid:"
                 << remotePid.toStdString().c_str();


    }

}

void Dispatcher::finalize()
{
    if (m_currentWidget) m_currentWidget->close();
    qApp->quit();
}

void Dispatcher::endIdle()
{
    QProcess p;
    p.start("sudo", QStringList() << "/opt/vc/bin/tvservice" << "--status");
    p.waitForFinished();
    QByteArray output = p.readAllStandardOutput();

    if (output.contains("[TV is off]"))
    {
        p.start("sudo", QStringList() << "/opt/vc/bin/tvservice" << "-p");
        p.waitForFinished();
        p.start("sudo", QStringList() << "chvt" << "1");
        p.waitForFinished();
        p.start("sudo", QStringList() << "chvt" << "7");
        p.waitForFinished();
    }
}

void Dispatcher::startInlineRemote()
{
    m_viewer->pause();
    QUrl remote_url(settings->get("remote/local_url").toString() +
         "?k=" + settings->get("remote/key").toString());
    qDebug() << remote_url;
    startBrowser(remote_url);
    Browser* browser = dynamic_cast<Browser*>(m_currentWidget);
    if (browser) browser->hideToolbar();
}
