#include <QEventLoop>
#include <QUrl>
#include <QDir>
#include <QTimer>
#include <QSettings>

#include <QDebug>

#include <unistd.h>

#include "downloader.h"

Downloader::Downloader(PlaylistLoop *loop, bool mode)
    : QObject()
    , m_loop(loop)
    , m_done(false)
    , m_updateMode(mode)
    , m_idl(NULL)
    , m_index(-1)
    , m_countError(0)
{
    m_baseUrl = loop->baseUrl();
    m_cacheDir = QDir::homePath() + "/.linutop-kiosk";
    m_dataDir = m_cacheDir + "/data";
    m_etagDir = m_cacheDir + "/etag";
    QDir d;
    d.mkpath(m_dataDir);
    d.mkpath(m_etagDir);
}

Downloader::~Downloader()
{
}

void Downloader::start()
{
    downloadNext();
}

void Downloader::addError()
{
    if(m_updateMode) m_updateMode = false;
    ++m_countError;
}

void Downloader::replaceFile()
{
    m_loop->items().replace(m_index, new PlaylistItem(
                                    "file://:/image/nothing_to_display.png",
                                    PlaylistItem::TypeUndefined, 10, false));
    downloadNext();
}

void Downloader::downloadNext()
{
    if (m_idl)
    {
        delete m_idl;
        m_idl = 0;
    }

    ++m_index;

    if (m_index >= m_loop->count())
    {
        QFile pl(m_cacheDir + "/playlist");
        QFile newpl(m_cacheDir + "/playlist-old");
        if(m_countError != 0)
        {
            if(newpl.open(QIODevice::ReadWrite) && pl.open(QIODevice::WriteOnly))
            {
                pl.write(newpl.readAll());
                newpl.close();
                pl.close();
            }
            qDebug() << "Errors occured while downloading.";
            Q_EMIT downloadNotDone();
        }
        else
        {
            m_done = true;
            newpl.remove();
            Q_EMIT downloadDone();
        }
        deleteLater();
        return;
    }

    qDebug() << "DL idx:" << m_index << "/" << m_loop->count() - 1;
    PlaylistItem *item = m_loop->at(m_index);

    // check if we need to download the data
    if (item->type() == PlaylistItem::TypeWebpage)
    {
        downloadNext();
        return;
    }

    QString itemUri = item->uri();
    if (itemUri.startsWith("file://"))
    {
        downloadNext();
        return;
    }

    if (itemUri.startsWith("rtsp://"))
    {
        downloadNext();
        return;
    }

    QStringList uri = item->uri().split('/');
    QFile m_file(uri.at(2));
    if(m_file.open(QIODevice::ReadOnly))
    {
        QString file = m_file.readAll();
        if(file.contains("can't find resource"))
        {
            qDebug() << "Can't find resource : nothing to display.";
            m_loop->items().replace(m_index, new PlaylistItem(
                                            "file://:/image/nothing_to_display.png",
                                            PlaylistItem::TypeUndefined, 10, false));
        }
    }
    m_file.close();


    QFile update_file(m_dataDir + "/maintenance.txt");
    if(update_file.exists())
    {
        if(item->type() != PlaylistItem::TypeMaintenance)
            update_file.remove();
        else if(update_file.open(QIODevice::ReadOnly))
        {
            QString file = update_file.readAll();
            if(file.contains("Update succeed!"))
            {

                qDebug() << "Maintenance done. Wait for new playlist.";
                m_loop->items().replace(m_index, new PlaylistItem(
                                                "file://:/image/nothing_to_display.png",
                                                PlaylistItem::TypeUndefined, 10, false));
            }
        }
    }

    QString filename(itemUri.toUtf8().toBase64().replace("/", "-"));
    QString localPath = m_dataDir + "/" + filename;
    QString localUri = QString("file://") + localPath;
    QString etagPath = m_etagDir + "/" + filename;
    item->setUri(localUri);

    if (itemUri.at(0) == '/')
        itemUri = m_baseUrl + "/r" + itemUri;

    // the ItemDownloader decides if it needs to use the cache or not
    m_idl = new ItemDownloader(itemUri, localPath, etagPath, m_updateMode);
    connect(m_idl, SIGNAL(finished()), this, SLOT(downloadNext()));
    connect(m_idl, SIGNAL(sizeIssue()), this, SLOT(replaceFile()));
    connect(m_idl, SIGNAL(onError()), this, SLOT(addError()));
    m_idl->start();
}


ItemDownloader::ItemDownloader(QString &uri, QString &dest, QString &etag, bool &mode)
    : QObject()
    , m_uri(uri)
    , m_update(mode)
    , m_coLost(false)
{
    m_file.setFileName(dest);
    m_etagFile.setFileName(etag);
    qDebug() << etag;
    m_nam = new QNetworkAccessManager();
    m_settings = Settings::load();
    m_cleanser = new Cleanser(m_settings);

}

void ItemDownloader::start()
{
    m_request = QNetworkRequest(QUrl(m_uri));

    if (inCache() && !m_uri.endsWith(".zip"))
    {
        qDebug() << "NOT downloading";
        Q_EMIT(finished());
        return;
    }

    qDebug() << "Downloading";
    QNetworkConfigurationManager manager;
    m_nam->setConfiguration(manager.defaultConfiguration());

    connect(m_nam, SIGNAL(networkAccessibleChanged(QNetworkAccessManager::NetworkAccessibility)),
            this, SLOT(namChanged(QNetworkAccessManager::NetworkAccessibility)));

    m_reply = m_nam->get(m_request);
    connect(m_reply, SIGNAL(downloadProgress(qint64, qint64 )),
            this, SLOT(checkSizeFileFirst(qint64, qint64 )));
}

bool ItemDownloader::inCache()
{


    if (! m_file.exists() || ! m_etagFile.exists())
    {
        return false;
    }
    m_etagFile.open(QIODevice::ReadOnly);
    QString localEtag(m_etagFile.readAll());
    QStringList fileSizeFromTag(localEtag.split(":"));
    if(m_file.size() != fileSizeFromTag.last().toLong() )
    {
        m_file.close();
        m_etagFile.close();
        m_file.remove();
        return false;
    }
    m_etagFile.close();

    if(!m_update)
        return true;

    QNetworkReply *reply = m_nam->head(m_request);

    // synchronous request
    QEventLoop loop;
    connect(reply, SIGNAL(finished()), &loop, SLOT(quit()));
    loop.exec();
    loop.deleteLater();



    if (reply->error() != QNetworkReply::NoError)
        return true;  // try to use the cache if we couldn't HEAD the file

    QString dateStr(reply->rawHeader("Last-Modified"));
    QString length(reply->rawHeader("Content-Length"));

    QString distantEtag = QString("%1:%2").arg(dateStr, length);

    if (distantEtag != localEtag)
        return false;

    return true;
}

ItemDownloader::~ItemDownloader()
{
    m_nam->deleteLater();
}

void ItemDownloader::checkSizeFileFirst(qint64 /*bytesRec*/, qint64 bytesTotal)
{
    pauseWhileCheckingSize();
    connect(m_cleanser, SIGNAL(ok()), this, SLOT(resumeDownload()));
    connect(m_cleanser, SIGNAL(notOk()), this, SLOT(notEnoughSpace()));
    m_cleanser->checkThat(bytesTotal);
}

void ItemDownloader::notEnoughSpace()
{
    qDebug() << "Not enough space...";
    Q_EMIT sizeIssue();
}

void ItemDownloader::pauseWhileCheckingSize()
{
    disconnect(m_reply, SIGNAL(downloadProgress(qint64, qint64 )),
            this, SLOT(checkSizeFileFirst(qint64, qint64 )));

    if(m_reply==0)
        return;

    m_reply->abort();
    m_reply = 0;
}

void ItemDownloader::resumeDownload()
{
    m_reply = m_nam->get(m_request);
    connect(m_reply, SIGNAL(finished()), this, SLOT(onFinished()));
    connect(m_reply, SIGNAL(readyRead()), this, SLOT(onReadyRead()));

    disconnect(m_cleanser, SIGNAL(ok()), this, SLOT(resumeDownload()));
    disconnect(m_cleanser, SIGNAL(notOk()), this, SLOT(notEnoughSpace()));
}

void ItemDownloader::namChanged(QNetworkAccessManager::NetworkAccessibility /*nam*/)
{
    qDebug() << "CONNECTION LOST";
    m_coLost = true;
    onFinished();
}

void ItemDownloader::onReadyRead()
{
    if (m_reply->error() != QNetworkReply::NoError)
        return;

    if (!m_file.isOpen())
        m_file.open(QIODevice::WriteOnly);

    m_file.write(m_reply->readAll());
}

void ItemDownloader::onFinished()
{
    if (m_reply->error() != QNetworkReply::NoError || m_coLost)
    {

        QFile tsFile(QDir::homePath() + "/.linutop-kiosk/playlist_last_timestamp.txt");
        QDir dataDir = QDir::homePath() + "/.linutop-kiosk/data";
        QDir etagDir = QDir::homePath() + "/.linutop-kiosk/etag";
        if(dataDir.count() == 0 || etagDir.count() == 0)
        {
            Q_EMIT(finished());
            qDebug() << "Network error, nothing in cache. Exit.";
            return;
        }


        if(tsFile.open(QIODevice::ReadWrite))
        {
            tsFile.resize(0);
            tsFile.close();
        }
        Q_EMIT(onError());
        Q_EMIT(finished());
        return;
    }
    m_file.write(m_reply->readAll());
    m_file.close();

    QString dateStr(m_reply->rawHeader("Last-Modified"));
    QString length(m_reply->rawHeader("Content-Length"));

    QString distantEtag = QString("%1:%2").arg(dateStr, length);
    if (m_etagFile.open(QIODevice::WriteOnly))
    {
        m_etagFile.write(distantEtag.toUtf8());
        m_etagFile.close();
    }

    Q_EMIT(finished());
}
