#include <QStringList>
#include <QFileInfo>
#include <QUrl>
#include <QNetworkAccessManager>
#include <QNetworkReply>
#include <QNetworkRequest>
#include <QEventLoop>
#include <QMimeDatabase>

#include <QDebug>

#include "playlistitem.h"

QMap<PlaylistItem::Type, QString> PlaylistItem::typeToStrMap(initStrings());

static PlaylistItem::Type guessType(QString uri)
{
    QUrl url(uri);

    if (url.isLocalFile())
    {
        QFileInfo info(url.path());
        if (info.isDir())
            return PlaylistItem::TypeDir;
    }

    if (url.scheme() == "rtsp")
        return PlaylistItem::TypeVideo;

    if (url.scheme() == "vlc")
        return PlaylistItem::TypeVideo;

    if(url.scheme() == "http" || url.scheme() == "https")
        return PlaylistItem::TypeWebpage;

    QMimeDatabase db;
    QMimeType mime = db.mimeTypeForUrl(url);
    qDebug() << "mime:" << mime.name();
    if (mime.name().startsWith("video/"))
        return PlaylistItem::TypeVideo;
    if (mime.name().startsWith("audio/"))
        return PlaylistItem::TypeVideo;
    if (mime.name().startsWith("image/gif"))
        return PlaylistItem::TypeGif;
    if (mime.name().startsWith("image/"))
        return PlaylistItem::TypeImage;
    if ((mime.name() == "text/html") || (mime.name() == "text/plain"))
        return PlaylistItem::TypeWebpage;
    if (mime.name() == "application/pdf")
        return PlaylistItem::TypePdf;
    if (mime.name() == "application/x-desktop" || mime.name() == "application/x-mswinurl")
        return PlaylistItem::TypeUrl;

    return PlaylistItem::TypeWebpage;
}

PlaylistItem::PlaylistItem(QString uri, PlaylistItem::Type type,
                           int time, bool href, QStringList param)
    : m_uri(uri)
    , m_type (type)
    , m_time(time)
    , m_href(href)
    , m_params(param)
{
    if (m_type == PlaylistItem::TypeUndefined)
    {
        qDebug() << "Guessing Type";
        m_type = guessType(m_uri);
    }

    if (m_type == PlaylistItem::TypeUrl)
    {
        qDebug() << "Converting in Url";
        QFile f(m_uri.mid(7));
        if(f.open(QIODevice::ReadOnly))
        {
            QString file = f.readAll();
            if(file.contains("URL="))
            {
                QStringList string = file.split("L=");
                m_uri = string.at(1);
                m_uri.remove("\n");
            }
        }
        f.close();
        m_type = guessType(m_uri);
    }
}

QMap<PlaylistItem::Type, QString> PlaylistItem::initStrings()
{
    QMap<PlaylistItem::Type, QString> m;

    m[PlaylistItem::TypeRemote] = "Remote";
    m[PlaylistItem::TypeVideo] = "Video";
    m[PlaylistItem::TypeGif] = "Gif";
    m[PlaylistItem::TypeImage] = "Image";
    m[PlaylistItem::TypeWebpage] = "Webpage";
    m[PlaylistItem::TypePdf] = "PDF";
    m[PlaylistItem::TypeDocument] = "Document";
    m[PlaylistItem::TypeIdle] = "Idle";
    m[PlaylistItem::TypeNotSupported] = "Not Supported";

    return m;
}

QString PlaylistItem::typeToStr(PlaylistItem::Type type)
{
    return typeToStrMap[type];
}

PlaylistItem *PlaylistItem::idle()
{
    //return new PlaylistItem(QString(), PlaylistItem::TypeIdle, 10);
    return new PlaylistItem(QString(), PlaylistItem::TypeIdle, 60);
}

PlaylistItem* PlaylistItem::fromXmlNode(QDomNode *node)
{
    QString uri, type_param;
    QStringList params;
    int time = 5;
    bool href = false;
    PlaylistItem::Type type = PlaylistItem::TypeUndefined;
    QDomNode tmp;

    QDomNamedNodeMap attributes = node->attributes();
    uri = attributes.namedItem("uri").nodeValue();

    tmp = attributes.namedItem("parameters");
    if (!tmp.isNull())
    {
        QString parameters = tmp.nodeValue();
        params = parameters.split(";");
        foreach (QString param, params) {
            if(param.contains("type:")){
                type_param = param.split(":").at(1);
            }
        }
    }

    tmp = attributes.namedItem("type");
    if (!tmp.isNull())
    {
        QString mediaType;
        if(type_param != "")
            mediaType = type_param;
        else
            mediaType = tmp.nodeValue();

        if(mediaType=="maintenance")
        {
            type = PlaylistItem::TypeMaintenance;
            return new PlaylistItem(uri, type, 0, false, params);
        }

        else if (mediaType == "image"){
            type = PlaylistItem::TypeImage;
            qDebug() << "image!!";
        }
        else if (mediaType == "audio")
            type = PlaylistItem::TypeVideo;
        else if (mediaType == "gif")
            type = PlaylistItem::TypeGif;
        else if (mediaType == "video")
            type = PlaylistItem::TypeVideo;
        else if (mediaType == "pdf")
            type = PlaylistItem::TypePdf;
        else if (mediaType == "document")
            type = PlaylistItem::TypeDocument;
        else if (mediaType == "url")
            type = PlaylistItem::TypeWebpage;
    }

    tmp = attributes.namedItem("time");
    if (!tmp.isNull())
        time = tmp.nodeValue().toInt();
    else if (type == PlaylistItem::TypeVideo)
        time = 0;

    tmp = attributes.namedItem("href");
    if (!tmp.isNull())
        href = (tmp.nodeValue() == "1");


    return new PlaylistItem(uri, type, time, href, params);
}
