#include <QStringList>
#include <QUrl>
#include <QProcess>
#include <QFileInfo>
#include <QDirIterator>
#include <qpixmap.h>

#include <QDebug>
#include <QLabel>

#include "playlistloop.h"
#include "playlistitem.h"
#include "settings.h"


PlaylistLoop::PlaylistLoop(int default_time, QDomNode &node, QString baseUrl)
    : m_baseUrl(baseUrl)
    , m_idle(false)
    , m_priority(0)
    , m_downloadable(true)
    , m_default_time(default_time)
    , m_shuffle(false)
{
    m_weekdays.resize(7);
    m_weekdays.fill(false);

    fromXmlNode(node);
    addDemoItems();
}

PlaylistLoop::PlaylistLoop(int default_time, QString &data, QString baseUrl)
    : m_baseUrl(baseUrl)
    , m_idle(false)
    , m_priority(0)
    , m_downloadable(false)
    , m_default_time(default_time)
    , m_shuffle(false)
{
    m_weekdays.resize(7);
    m_weekdays.fill(false);

    fromString(data);
    addDemoItems();
}

PlaylistLoop::~PlaylistLoop()
{
    //delete m_items;
}

namespace detail {

struct PlaylistItemCompare
{
    bool operator ()(PlaylistItem *a, PlaylistItem *b)
    {
        QString a_low = a->uri().toLower(), b_low = b->uri().toLower();
        if (a_low == b_low)
            return a->uri() < b->uri();
        return a_low < b_low;
    }
};

} // detail

QDateTime PlaylistLoop::getFromIdle(QString uri)
{
    QStringList from_l = uri.mid(uri.indexOf("From:")+6,5).split(":");
    QTime from_t;

    from_t.setHMS(from_l.at(0).toInt(),from_l.at(1).toInt(),0,0);
    from.setTime(from_t);
    from.setDate(QDate::currentDate());
    return from;
}

QDateTime PlaylistLoop::getToIdle(QString uri)
{
    QStringList to_s = uri.mid(uri.indexOf("To:")+4,5).split(":");
    QTime to_t;

    to_t.setHMS(to_s.at(0).toInt(),to_s.at(1).toInt(),0,0);
    to.setTime(to_t);
    to.setDate(QDate::currentDate());
    return to;
}

void PlaylistLoop::fromString(QString &data)
{
    int lastSpace;
    int time = 0;
    QString uri;
    QStringList lines = data.split("\n");

    Settings* settings = Settings::load();

    // Add drag and drop directory if it contains elements and remote activated.
    if (settings->get("remote/enabled").toBool())
    {
        QString drag_drop_dir = qgetenv("HOME") + "/.linutop-kiosk/dragndrop";
        if (QDir(drag_drop_dir).exists())
        {
            qDebug() << "Add drag and drop directory.";
            addDirectoryInItems(drag_drop_dir, m_default_time);
        }
    }

    bool usb_auto_ok = true;
    bool has_other_than_usb_auto = false;
    int nb_webpage = 0;
    for (int i = 0; i < lines.count(); ++i)
    {
        QString line = lines.at(i);

        // Skip commented lines.
        if (line.startsWith("#")) continue;

        // Skip bad formatted lines.
        {
            lastSpace = line.lastIndexOf(' ');
            if (lastSpace < 0) continue;

            bool ok = false;
            time = line.mid(lastSpace).toInt(&ok);
            if (!ok) continue;
        }

        uri = line.left(lastSpace);

        if(uri.contains("Screen OFF")){
            m_items.append(new PlaylistItem(
                               uri, PlaylistItem::TypeIdle,
                               0, false));
        }

        // USB auto.
        if (uri == "usb-auto")
        {
            qDebug() << "USB Auto in playlist.";
            usb_auto_ok = addDirectoryInItems(
                        qgetenv("HOME") + "/.linutop-kiosk/usbauto", time);
            continue;
        }

        QUrl url(uri);
        if (!url.isValid()) continue;

        if (url.scheme() == "http" || url.scheme() == "https")
        {
            nb_webpage++;
        }

        // Local file.
        if (url.scheme() == "file")
        {
            QFileInfo info(url.path());
            if (!info.exists())
                continue;
            if (!info.isDir())
            {
                m_items.append(new PlaylistItem(
                                   uri, PlaylistItem::TypeUndefined,
                                   time, false));
            }
            else
                addDirectoryInItems(url.path(), time);
        }

        // Samba file.
        else if (url.scheme() == "smb")
        {
            QString fullpath;
            {
                qDebug() << "Receiving smb url:" << url.toEncoded();

                QString localUrl("%0/gvfs/smb-share:server=%1,share=%2/");
                {
                    QStringList split = QString(url.toEncoded()).split('/');
                    qDebug() << "  Server:" << split.at(2);
                    qDebug() << "  Share :" << split.at(3);
                    localUrl = localUrl.arg(
                                qgetenv("XDG_RUNTIME_DIR"),
                                split.at(2), split.at(3));
                    qDebug() << "  Mount url:" << localUrl;

                    // Test if already mounted, otherwise mount it.
                    {
                        QFileInfo info(localUrl);
                        if (!info.exists())
                        {
                            QProcess gvfs;
                            gvfs.start("gvfs-mount",
                                       QStringList()
                                       << QString("smb://%1/%2")
                                       .arg(split.at(2), split.at(3)));
                            gvfs.waitForFinished();
                        }
                    }
                }

                QString relative;
                {
                    QStringList split = QString(url.toString()).split('/');
                    split.pop_front();
                    split.pop_front();
                    split.pop_front();
                    split.pop_front();
                    relative = split.join("/");
                }

                fullpath = localUrl+relative;
                qDebug() << "  Fullpath :" << fullpath;
            }

            QFileInfo info(fullpath);
            if (!info.exists())
                continue;
            if (!info.isDir())
            {
                m_items.append(new PlaylistItem(
                                   fullpath, PlaylistItem::TypeUndefined,
                                   time, false));
            }
            else
                addDirectoryInItems(fullpath, time);
        }

        // Otherwise.
        else
        {
            if(lines.count() -2 == nb_webpage)
                // In case the playlist is full of webpages (to handle network error...see webpage.cpp)
            {
                m_items.append(new PlaylistItem(
                            uri, PlaylistItem::TypeUndefined, time, false, QStringList("all_webpages")));
            }
            else {
                m_items.append(new PlaylistItem(
                            uri, PlaylistItem::TypeUndefined, time, false));
            }
        }

        has_other_than_usb_auto = true;
    }

    // Warn if auto-usb, but no file present...
    if (!usb_auto_ok && !has_other_than_usb_auto)
    {
        qDebug() << "Dispay warning for usb-auto.";
        m_items.append(new PlaylistItem(
                           "file://:/image/usb_auto_insert_warning.png",
                           PlaylistItem::TypeUndefined, time, false));
    }
    if(settings->get("slideshow/shuffle").toBool())
        std::random_shuffle(m_items.begin(), m_items.end());
}

void PlaylistLoop::fromXmlNode(QDomNode &node)
{
    QDomNode tmp;
    QDomNamedNodeMap attributes = node.attributes();

    tmp = attributes.namedItem("idle");
    if (!tmp.isNull())
        m_idle = (tmp.nodeValue() == "1");
    tmp = attributes.namedItem("shuffle");
    if (!tmp.isNull())
        m_shuffle = (tmp.nodeValue()=="1");
    tmp = attributes.namedItem("start");
    if (!tmp.isNull())
        m_startTimeStr = tmp.nodeValue();
    tmp = attributes.namedItem("end");
    if (!tmp.isNull())
        m_endTimeStr = tmp.nodeValue();
    tmp = attributes.namedItem("repeat");
    if (!tmp.isNull())
        m_weekdaysStr = tmp.nodeValue();
    tmp = attributes.namedItem("from");
    if (!tmp.isNull())
        m_startDateStr = tmp.nodeValue();
    tmp = attributes.namedItem("to");
    if (!tmp.isNull())
        m_endDateStr = tmp.nodeValue();
    tmp = attributes.namedItem("id");
    if (!tmp.isNull())
        m_id = tmp.nodeValue();

    if (!validate())
        return;

    /* idle loops don't have items, we create one */
    if (m_idle)
    {
        m_downloadable = false;
        PlaylistItem * item = PlaylistItem::idle();
        m_items.append(item);
    }
    else
    {
        QDomNodeList items = node.childNodes();
        int nb_webpages = 0;
        for (int i=0; i<items.count(); ++i)
        {
            QDomNode n = items.at(i);
            PlaylistItem * item = PlaylistItem::fromXmlNode(&n);
            if (item->isEmpty())
                continue;

            if(item->type() == 6)
                nb_webpages++;

            if(nb_webpages == items.count())
            {
                QStringList params = item->params();
                params.append("all_webpages");
                item = new PlaylistItem(item->uri(), item->type(),
                                        item->time(), item->href(),
                                        params);

            }
            m_items.append(item);

        }
        if(m_shuffle)
            std::random_shuffle(m_items.begin(), m_items.end());
    }
}

bool PlaylistLoop::addDirectoryInItems(const QString &path, int time)
{
    // QDirIterator is a random iterator.
    // That's why we need to sort the list at the end
    // of the else statement.
    QDirIterator iterator(path, QDirIterator::Subdirectories);
    QList<PlaylistItem*> item_children;

    bool ok = false;
    while (iterator.hasNext())
    {
        iterator.next();
        if (iterator.fileInfo().isDir())
            continue;

        PlaylistItem * item = new PlaylistItem(
                    QString("file://") + iterator.filePath(),
                    PlaylistItem::TypeUndefined, time, false);
        if (item->type() == PlaylistItem::TypeVideo)
            item->setTime(0);
        item_children.append(item);
        ok = true;
    }

    // Sort the list.
    std::sort(item_children.begin(), item_children.end(),
              detail::PlaylistItemCompare());

    // Append the sorted list.
    m_items.append(item_children);

    return ok;
}

bool PlaylistLoop::validate()
{

    if (! validateDate())
      return false;
    if (! validateTime())
        return false;
    if (! validateWeekdays())
        return false;


    return true;
}

bool PlaylistLoop::validateDate()
{
    bool sWasEmpty = false;
    bool eWasEmpty = false;
    if (m_startDateStr.isEmpty())
    {
        m_startDateStr = QDate::currentDate().toString("yyyy/MM/dd");
        sWasEmpty = true;
    }
    if (m_endDateStr.isEmpty())
    {
        m_endDateStr = QDate::currentDate().toString("yyyy/MM/dd");
        eWasEmpty = true;
    }


    QStringList spl;
    QDate start_d, end_d;

    spl = m_startDateStr.split("/");
    if (spl.count() != 3)
        return false;
    start_d = QDate(spl.at(0).toInt(), spl.at(1).toInt(), spl.at(2).toInt());
    if (!start_d.isValid())
        return false;

    spl = m_endDateStr.split("/");
    if (spl.count() != 3)
        return false;
    end_d = QDate(spl.at(0).toInt(), spl.at(1).toInt(), spl.at(2).toInt());
    if (!end_d.isValid())
        return false;

    if (end_d < start_d)
        return false;

    if(sWasEmpty)
        m_startDateStr = "";
    if(eWasEmpty)
        m_endDateStr = "";

    m_startDate.setDate(start_d);
    m_endDate.setDate(end_d);

    return true;
}

bool PlaylistLoop::validateTime()
{
    QStringList spl;
    int s_h, s_m, e_h, e_m, e_mm;

    s_h = s_m = e_h = e_m = e_mm = 0;

    spl = m_startTimeStr.split(":");
    if (spl.count() != 2)
    {
        s_h = 0;
        s_m = 0;
        m_priority -= 1;
    }
    else
    {
        s_h = spl.at(0).toInt();
        s_m = spl.at(1).toInt();
        if ((s_h < 0) || (s_h > 23) || (s_m < 0) || (s_m > 59))
        {
            s_h = 0;
            s_m = 0;
        }
    }

    spl = m_endTimeStr.split(":");
    if (spl.count() != 2)
    {
        e_h = 23;
        e_m = 59;
        m_priority -= 1;
    }
    else
    {
        e_h = spl.at(0).toInt();
        e_m = spl.at(1).toInt();
        if ((e_h < 0) || (e_h > 23) || (e_m < 0) || (e_m > 59))
        {
            e_h = 23;
            e_m = 59;
        }
    }

    if(s_h == e_h && s_m == e_m)
    {
        m_endDate.setDate(QDate::currentDate());
    }

    m_startDate.setTime(QTime(s_h, s_m));
    m_endDate.setTime(QTime (e_h, e_m, e_mm));
    if (!m_startDate.isValid() || !m_endDate.isValid())
        return false;

    return true;
}

bool PlaylistLoop::validateWeekdays()
{
    if (m_weekdaysStr.isEmpty())
    {
        m_priority -= 1;
        m_weekdays.fill(true);
        return true;
    }

    for (int i=0; i < m_weekdaysStr.count(); ++i)
    {
        int l = m_weekdaysStr.at(i).digitValue();
        if ((l >= 0) && (l < 7))
            m_weekdays.setBit(l, true);
    }

    return true;
}

void PlaylistLoop::addDemoItems()
{

#ifdef DEMO
    for(int i = 1;i<m_items.count();i+=4) {
        m_items.insert(i,new PlaylistItem("file://:/demo/demo.jpg",PlaylistItem::TypeImage,3));
    }
#endif //DEMO
}

QDateTime PlaylistLoop::nowOk(QDateTime start_sec)
/*start_sec = current time at the begining of the next item*/
{
    QDateTime now = QDateTime::currentDateTime();
    QDateTime out;

    if(!validate())
        return out;

    //Au lancement, permet de faire un ajustement sur la date si le jour de début
    // ou de fin n'est pas renseigné
    if(m_startDate.date() == m_endDate.date() && m_startDate.time() > m_endDate.time())
    {
        QTime curTime =  QTime::currentTime();
        if(curTime >= QTime::fromString("00:00:00", "HH:mm:ss")
                && curTime < m_endDate.time())
            m_startDate = m_startDate.addDays(-1);
        else
            m_endDate = m_endDate.addDays(1);
    }

    //qDebug() << id() << m_startDate << m_endDate;
    int weekday = now.date().dayOfWeek() - 1; /* monday == 0 for us, 1 for Qt */


    if (!m_idle && m_items.count() == 0)
        return out;

    if (! m_weekdays.at(weekday))
        return out;

    if (m_startDate.time() == m_endDate.time())
    {
        if (now < m_startDate)
            return out;

        if (start_sec > m_startDate)
            return out;
    }

    if(m_startDate.time() != m_endDate.time())
    {
        if((now < m_startDate) || (now > m_endDate))
            return out;
    }

    return m_startDate;
}
