#include <cassert>

#include <QWebSettings>
#include <QSqlDatabase>
#include <QSettings>
#include <QShortcut>
#include <QEventLoop>
#include <QDesktopWidget>
#include <QWebFrame>
#include <QWebElement>
#include <QCursor>

#include <QDebug>

#include "ui/exitpassworddialog.h"
#include "ui/widgets/browser/browser.h"
#include "ui/widgets/webpage.h"
#include "viewer.h"
#include "utils.h"

PlaylistEventFilter::PlaylistEventFilter()
    : QObject()
    , m_paused(true)
{
    settings = Settings::load();
}

bool PlaylistEventFilter::eventFilter(QObject *obj, QEvent *event)
{
    if (m_paused) {
        if (event->type() == QEvent::MouseButtonPress) {
            return true;
        }
        return false;
    }

    switch (event->type())
    {
        case QEvent::MouseMove: {
            if (settings->get("main/mode").toString() == "kiosk")
            {
                qobject_cast<Viewer *>(obj->parent())->emitMouseEvent();
                return true;
            }
            break;
        }
        case QEvent::MouseButtonPress: {
            qobject_cast<Viewer *>(obj->parent())->mousePressEvent((QMouseEvent *)event);
            return true;
        }
        case QEvent::KeyPress:
            qobject_cast<QWidget *>(obj)->setCursor(Qt::BlankCursor);
            /* hum, not nice... */

            return true;
        default: break;
    }

    return false;
}

bool BrowserEventFilter::eventFilter(QObject *obj, QEvent *event)
{
    QMouseEvent *me = dynamic_cast<QMouseEvent *>(event);
    QInputEvent *ie = dynamic_cast<QInputEvent *>(event);
    if (me || ie)
        qobject_cast<Viewer *>(obj->parent())->resetTimer();

    return false;
}

Viewer::Viewer() : QStackedWidget()
{
    settings = Settings::load();

    QWidget::connect(&m_timer, SIGNAL(timeout()), this, SLOT(onTimerTimeout()));

    m_playlistEventFilter = new PlaylistEventFilter;
    m_browserEventFilter = new BrowserEventFilter;
    m_unlockTimer.setSingleShot(true);
    QWidget::connect(&m_unlockTimer, SIGNAL(timeout()),
            this, SLOT(unlockPlaylistEventFilter()));
    QWidget::connect(this, SIGNAL(currentChanged(int)), this, SLOT(on_currentChanged(int)));

    loadWindowSettings();
    setShortcuts();

    // Security setup.
    {
        bool fullscreen = settings->get("main/fullscreen").toBool();
        bool use_password = settings->get("main/admin_passwd").toBool();

        int largeur = settings->get("main/largeur_edit").toInt();
        int hauteur = settings->get("main/hauteur_edit").toInt();

        bool fixedscreen = settings->get("main/fixedscreen").toBool();

        QScopedPointer<panel::Manager> tmp1(
            panel::Manager::create(fullscreen, use_password));
        m_panel_manager.swap(tmp1);

        QScopedPointer<SystemShortcutManager> tmp2(
            new SystemShortcutManager(fullscreen, use_password));
        m_system_shortcut_manager.swap(tmp2);

        if (fullscreen && !fixedscreen){
           showFullScreen();
        }
        else if (!fullscreen && fixedscreen)
        {
            setGeometry(0,0,largeur,hauteur);
            setWindowFlags(Qt::FramelessWindowHint);
            show();
        }
        else if (fullscreen && fixedscreen)
        {
            showFullScreen();
        }
        else if (!fullscreen && !fixedscreen)
        {
            showNormal();
        }
    }

    // Hide mouse.
    if (settings->get("slideshow/hide_pointer").toBool())
    {
        qDebug() << "Hiding mouse pointer.";
        QScopedPointer<QProcess> hide(new QProcess);
        hide->start("unclutter", QStringList() << "-idle" << "1");
        m_hideMouse.swap(hide);
    }

    // Usb key detection setup.
    {
        if (settings->get("slideshow/usb_auto").toBool() &&
            (settings->get("main/mode").toString() == "display"))
        {
            QString remoteKey = settings->get("remote/key").toString();
            QScopedPointer<UsbAuto> usbauto(new UsbAuto(remoteKey));
            m_usbauto.swap(usbauto);
            QObject::connect(m_usbauto.data(), SIGNAL(finished()),
                             this, SLOT(requestRestart()));
        }
    }

    // Mute setting.
    {
        QStringList outputList;
        #ifndef __arm__
        outputList << "Master" << "Headphone" << "Speaker" << "PCM";
        #else
        outputList << "PCM";
        #endif

        if (settings->get("slideshow/mute").toBool())
        {
            Q_FOREACH(QString output, outputList)
            {
                qDebug() << "Disabling sound on channel:" << output;
                QStringList args;
                args << "set" << output << "mute";
                QProcess mute;
                mute.start("amixer", args);
                mute.waitForFinished();
            }
        }
        else
        {
            Q_FOREACH(QString output, outputList)
            {
                qDebug() << "Enabling sound on channel:" << output;
                QStringList args;
                args << "set" << output << "unmute";
                QProcess mute;
                mute.start("amixer", args);
                mute.waitForFinished();
            }
        }

    }

    // Web browser setup.
    {
        QWebSettings *websettings = QWebSettings::globalSettings();
        websettings->setAttribute(QWebSettings::JavascriptCanOpenWindows, true);
        websettings->setAttribute(QWebSettings::PluginsEnabled, true);
        websettings->setAttribute(QWebSettings::JavaEnabled, true);



        QString fontStr = settings->get("webpage/default_font").toString();
        if (!fontStr.isEmpty())
        {
            QFont font; font.fromString(fontStr);
            websettings->setFontFamily(QWebSettings::StandardFont, font.family());
            websettings->setFontSize(QWebSettings::DefaultFontSize, font.pointSize());
        }
    }

    // Cookie database setup ?
    {
        QSqlDatabase db = QSqlDatabase::addDatabase("QSQLITE");
        db.setDatabaseName(settings->cookiesFilename());
        db.open();
    }
}

Viewer::~Viewer(void) {}

void Viewer::emitMouseEvent()
{
    Q_EMIT mouseEvent();
}

void Viewer::mousePressEvent(QMouseEvent *event)
{
    switch(event->button()){
        case Qt::RightButton:
            if (settings->get("remote/enabled").toBool()
                    && settings->get("main/mode").toString() == "display")
            {
                killExternalPlayer();
                Q_EMIT startInlineRemote();
            }
            break;
        case Qt::MiddleButton:
            Q_EMIT middleclick();
            break;
        case Qt::LeftButton:
            if(settings->get("main/mode").toString() == "display")
                Q_EMIT leftclick();
            break;
        default: break;
    }
}

void Viewer::loadWindowSettings()
{
    QSettings appsettings;
    appsettings.beginGroup("Viewer");
    resize(appsettings.value("size", QSize(800, 600)).toSize());
    move(appsettings.value("pos", QPoint(200, 200)).toPoint());
    appsettings.endGroup();
}

void Viewer::saveWindowSettings()
{
    QSettings appsettings;
    appsettings.beginGroup("Viewer");
    appsettings.setValue("size", size());
    appsettings.setValue("pos", pos());
    appsettings.endGroup();
}

void Viewer::setShortcuts()
{
    QShortcut *shortcut;

    shortcut = new QShortcut(QKeySequence(Qt::CTRL + Qt::Key_Q), this);
    shortcut->setContext(Qt::ApplicationShortcut);
    connect(shortcut, SIGNAL(activated()), this, SLOT(bye()));

    if (!settings->get("main/admin_passwd").toBool())
    {
        shortcut = new QShortcut(QKeySequence(Qt::Key_Escape), this);
        shortcut->setContext(Qt::ApplicationShortcut);
        connect(shortcut, SIGNAL(activated()), this, SLOT(bye()));
    }
}

void Viewer::bye()
{
    if (settings->get("main/admin_passwd").toBool())
    {
        ExitPasswordDialog d;
        QString password = d.run();
        if (password.isEmpty() || ! Utils::validateAdminPassword(password))
            return;
    }

    if (m_prevWidget)
    {
        Runnable *r = dynamic_cast<Runnable *>(m_prevWidget);
        if (r)
            r->stop();
    }
    saveWindowSettings();
    Q_EMIT terminate();
}

void Viewer::onTimerTimeout()
{
    m_timer.stop();
    Q_EMIT end();
}

void Viewer::stop() { m_timer.stop(); }

void Viewer::setTimer()
{
    m_timer.setInterval(settings->get("main/stop_session_after_seconds").toInt() * 1000);
    m_timer.start();
}

void Viewer::displayWebpage(Webpage *page)
{
    // Display webapage.
    {
        addWidget(page);
        QObject::connect(page, SIGNAL(end()), this, SLOT(onWidgetTimeout()));
        QObject::connect(page, SIGNAL(middleclick()), this, SLOT(bye()));
        page->show();
        setCurrentWidget(page);
    }

    // Stop running widget.
    Runnable *r = dynamic_cast<Runnable*>(m_prevWidget);
    if (r && r->isActive()) this->pause();

    // Take screenshot.
    //Q_EMIT readyToShoot();
}

void Viewer::on_currentChanged(int /*index*/)
{
    QTimer::singleShot(100, this, SLOT(takeScreen()));
}

void Viewer::takeScreen()
{
    Q_EMIT readyToShoot();
}

void Viewer::resetTimer()
{
    if (this->m_timer.isActive()) {
        m_timer.start();
    }
}

void Viewer::closeEvent(QCloseEvent *event)
{
    bye();
    event->ignore();
}

bool Viewer::next(QWidget *widget)
{
    if (!widget) return false;

    addWidget(widget);
    widget->setMouseTracking(true);

    Browser *b = qobject_cast<Browser *>(widget);
    if (b)
    {
        if (settings->get("main/stop_session_after_seconds").toInt())
        {
            widget->installEventFilter(m_browserEventFilter);

            // Special case if the current page is the remote.
            if (b->getHome().toString() !=
                    settings->get("remote/local_url").toString() +
                    "?k=" + settings->get("remote/key").toString())
            {
                connect(widget, SIGNAL(mouseMoved()), this, SLOT(resetTimer()));
                setTimer();
            }
        }
    }
    else
    {
        widget->setCursor(Qt::BlankCursor);

        #ifndef __arm__
            QCursor curs = widget->cursor();
            if(settings->get("slideshow/external_player").toBool() ||
                    settings->get("slideshow/external_browser").toBool())
                curs.setPos(4000,4000);
        #endif
        if (m_playlistEventFilter &&
            settings->get("slideshow/stop_on_event").toBool())
            widget->installEventFilter(m_playlistEventFilter);
    }

    QObject::connect(widget, SIGNAL(end()), this, SLOT(onWidgetTimeout()));
    if(settings->get("main/mode").toString() == "display")
    {
        if(!settings->get("main/admin_passwd").toBool())
            QObject::connect(this, SIGNAL(middleclick()), this, SLOT(bye()));

    }

    /* VLC send a MouseMove event that might stop the slideshow, so temporarily
     * ignore the events
     */
    if (m_playlistEventFilter)
        m_playlistEventFilter->setPaused(true);

    m_unlockTimer.start(1000);
    setCurrentWidget(widget);
    QPalette plt;
    plt.setColor(QPalette::Window, Qt::black);
    widget->setPalette(plt);
    widget->setAutoFillBackground(true);
    widget->show();

    bool started = dynamic_cast<Runnable *>(widget)->start();
    if (!started) return false;

    if (m_prevWidget)
    {
        Runnable* run = dynamic_cast<Runnable*>(m_prevWidget);
        if (run) run->stop();
        removeWidget(m_prevWidget);
        m_prevWidget->setParent(NULL);
        delete m_prevWidget;
    }

    m_prevWidget = widget;

    return true;
}

void Viewer::pause()
{
    Runnable *r = dynamic_cast<Runnable *>(m_prevWidget);
    if (r) r->pause();
}

void Viewer::showText(const QString &text, int duration)
{
    // Create webpage to display text.
    Webpage* page;
    {
        qDebug() << "Displaying webpage:"
                 << QUrl::fromLocalFile(
                        "/usr/local/share/linutop-webkiosk/text.html")
                    .toString();
        qDebug() << "For" << duration << "seconds.";

        PlaylistItem* item = new PlaylistItem(
                    "file:///usr/share/linutop-webkiosk/text.html",
                    PlaylistItem::TypeWebpage,
                    duration);

        page = new Webpage(item, this);
        page->start();

        QEventLoop loop;
        QObject::connect(page, SIGNAL(loadFinished(bool)), &loop, SLOT(quit()));
        loop.exec();
        loop.deleteLater();

        page->page()->mainFrame()->findFirstElement("#text").setPlainText(text);
    }

    displayWebpage(page);
}

void Viewer::showUrl(QString url, int duration)
{
    // Create webpage to display text.
    Webpage* page;
    {
        qDebug() << "Displaying webpage:" << QUrl(url).toString();
        qDebug() << "For" << duration << "seconds.";

        PlaylistItem* item = new PlaylistItem(
                    QUrl(url).toString(),
                    PlaylistItem::TypeWebpage,
                    duration);

        page = new Webpage(item, this);
        page->start();
    }

    displayWebpage(page);
}

void Viewer::unlockPlaylistEventFilter()
{
    m_playlistEventFilter->setPaused(false);
}

void Viewer::onWidgetTimeout()
{
    m_timer.stop();
    Q_EMIT end();
}

void Viewer::killExternalPlayer()
{
    QProcess p;
    QString player;
#if __arm__
    player = "omxplayer";
#else
    player = settings->get("slideshow/external_player_cmd").toString();
    if(!player.isEmpty())
    {
        QStringList cmd = player.split(" ");
        player = cmd.at(0);
    }
    else
        player = "mplayer";
#endif
    p.start("pkill -9 " + player);
    p.waitForFinished();
}

void Viewer::requestRestart()
{Q_EMIT restart(); }
