#include <unistd.h>
#include <sys/types.h>

#include <QDir>
#include <QAction>
#include <QTabBar>
#include <QPrinter>
#include <QPrintDialog>
#include <QMouseEvent>

#include "restrictedwebpage.h"
#include "browser.h"


Browser::Browser(QWidget *parent)
    : QWidget(parent)
    , Runnable(-1)
{
    settings = Settings::load();

    QString tmpPath = QString("/tmp/webkiosk-%1").arg(getuid());
    QDir dir(tmpPath);
    dir.mkpath(tmpPath);

    m_processManager.setParent(this);

    m_tabWidget.show();
    //m_statusBar.show();
    m_layout.setContentsMargins(0, 0, 0, 0);
    m_layout.addWidget(&m_tabWidget);
    //m_layout.addWidget(&m_statusBar);
    setLayout(&m_layout);

    setShortcuts();
    setHome(settings->get("main/default_url").toString());
    m_newTabUrl = settings->get("main/new_tab_url").toString();
    if (m_newTabUrl.isEmpty())
        m_newTabUrl = m_home;

    m_tabWidget.setTabsClosable(true);
    m_tabWidget.setMovable(true);
    m_tabWidget.setElideMode(Qt::ElideMiddle);
    connect(&m_tabWidget, SIGNAL(tabCloseRequested(int)),
            this, SLOT(onTabCloseRequested(int)));

    if (! settings->get("main/hide_toolbar").toBool())
    {
        m_cornerBtn.setParent(this);
        m_cornerBtn.setIcon(QIcon::fromTheme("list-add"));
        m_cornerBtn.show();
        m_tabWidget.setCornerWidget(&m_cornerBtn, Qt::TopLeftCorner);
        connect(&m_cornerBtn, SIGNAL(clicked()), this, SLOT(newPage()));
    }

    downloads = 0;
    m_downloadLabel.setPixmap(QIcon::fromTheme("download").pixmap(QSize(24, 24)));
    m_downloadLabel.hide();
    m_tabWidget.setCornerWidget(&m_downloadLabel, Qt::TopRightCorner);

    remaining = 0;
    if (settings->get("main/with_timer").toBool())
    {
        remaining = settings->get("main/default_timer").toInt();
        setTimeout(1);
        m_statusBar.insertPermanentWidget(1, &m_timerLabel);
    }

    setContextMenuPolicy(Qt::CustomContextMenu);
}


Browser::~Browser(void) {}


QUrl Browser::getHome(void)
{ return QUrl(m_home); }


void Browser::setShortcuts(void)
{
    QAction *action;

    if (! settings->get("main/hide_toolbar").toBool())
    {
        action = new QAction(this);
        action->setShortcuts( QList<QKeySequence>() << Qt::CTRL + Qt::Key_L);
        connect(action, SIGNAL(triggered()), this, SLOT(selectAddressBar()));
        addAction(action);

        action = new QAction(this);
        action->setShortcuts( QList<QKeySequence>() << Qt::CTRL + Qt::Key_K);
        connect(action, SIGNAL(triggered()), this, SLOT(selectAddressBarAndBeginSearch()));
        addAction(action);

        action = new QAction(this);
        action->setShortcuts( QList<QKeySequence>() << Qt::CTRL + Qt::Key_T);
        connect(action, SIGNAL(triggered()), this, SLOT(newPage()));
        addAction(action);

        action = new QAction(this);
        action->setShortcuts( QList<QKeySequence>() << Qt::CTRL + Qt::Key_W);
        connect(action, SIGNAL(triggered()), this, SLOT(closeCurrentTab()));
        addAction(action);

        action = new QAction(this);
        action->setShortcuts( QList<QKeySequence>() << Qt::CTRL + Qt::Key_PageUp);
        connect(action, SIGNAL(triggered()), this, SLOT(nextTab()));
        addAction(action);

        action = new QAction(this);
        action->setShortcuts( QList<QKeySequence>() << Qt::CTRL + Qt::Key_PageDown);
        connect(action, SIGNAL(triggered()), this, SLOT(prevTab()));
        addAction(action);
    }
}


void Browser::setHome(QString url)
{
#ifdef DEMO
    url = "http://linutop.com";
#endif
    m_home = QUrl(url);
}


bool Browser::startImpl(void)
{ return newPage(m_home); }


bool Browser::stopImpl(void)
{
    m_processManager.empty();
    QString cmd = QString("rm -rf /tmp/webkiosk-%1").arg(getuid());
    int ret = system(cmd.toUtf8().data());
    if (ret != 0) {
        qDebug() << "Execution failed for:" << cmd << "->" << ret;
    }
    return true;
}


bool Browser::onTimeoutImpl(void)
{
    remaining -= 1;
    int hours = remaining / 3600;
    int minutes = (remaining - (hours * 3600)) / 60;
    int seconds = remaining - (hours * 3600 + minutes * 60);
    QString t;
    t.sprintf("remaining: %02d:%02d:%02d", hours, minutes, seconds);
    m_timerLabel.setText(t);

    if (remaining == 0)
    {
        Q_EMIT end();
        return true;
    }

    return false;
}


void Browser::pauseImpl(void) {}


void Browser::resumeImpl(void) {}


void Browser::onTabCloseRequested(int index)
{
    if (m_tabWidget.count() < 2) return;

    m_tabWidget.removeTab(index);
    Page *w = m_pages.takeAt(index);
    w->removeEmbedElements();
    delete w;
    refreshTabBar();
}


void Browser::closeCurrentTab(void)
{
    int idx = m_tabWidget.currentIndex();
    onTabCloseRequested(idx);
}


Page *Browser::newPage(RestrictedWebView *webview)
{
    Page *p = new Page(webview);
    setupPage(p);
    return p;
}


Page *Browser::newPage(QUrl url)
{
    if (url.isEmpty())
        url = m_newTabUrl;
    Page *p = new Page(url);

    setupPage(p, url);
    return p;
}


void Browser::setupPage(Page *p, QUrl url)
{
    QString url_s("New Page");
    if (url.isValid())
        url_s = url.toString();

    m_tabWidget.addTab(p, url_s);
    m_pages << p;
    m_tabWidget.setCurrentWidget(p);

    QObject::connect(
                p, SIGNAL(titleChanged(QString &, Page *)),
                this, SLOT(onTitleChanged(QString &, Page *)));
    QObject::connect(
                p->view(), SIGNAL(mouseMoved()),
                this, SLOT(onMouseMove()));
    QObject::connect(
                p, SIGNAL(end()),
                this, SLOT(onEnd()));
    QObject::connect(
                qobject_cast<RestrictedWebView *>(p->view()),
                SIGNAL(webViewCreated(RestrictedWebView *)),
                this, SLOT(onWebViewCreated(RestrictedWebView *)));
    QObject::connect(
                qobject_cast<RestrictedWebPage *>(p->view()->page()),
                SIGNAL(downloadStarted()),
                this, SLOT(onDownloadStarted()));
    QObject::connect(
                qobject_cast<RestrictedWebPage *>(p->view()->page()),
                SIGNAL(downloadDone(QString &, QString &)),
                this, SLOT(onDownloadDone(QString &, QString &)));
    QObject::connect(
                p, SIGNAL(urlDenied(QString &)),
                this, SLOT(onUrlDenied(QString &)));
    QObject::connect(
                p->navbar(), SIGNAL(printRequested()),
                this, SLOT(onPrintRequested()));

    refreshTabBar();
}


void Browser::refreshTabBar()
{
    if (! settings->get("main/hide_toolbar").toBool())
        return;

    /* if the toolbar is hidden, only show the tabbar if we have > 1 tabs */
    if (m_tabWidget.count() > 1)
        m_tabWidget.tabBar()->show();
    else
        m_tabWidget.tabBar()->hide();
}


void Browser::onPrintRequested()
{
    QPrinter printer;
    QPrintDialog dialog(&printer, this);
    QString p_name = settings->get("main/printer_name").toString();
    if (! p_name.isEmpty())
        printer.setPrinterName(p_name);
    Page *page = qobject_cast<Page *>(m_tabWidget.currentWidget());
    page->view()->print(&printer);
}


void Browser::onWebViewCreated(RestrictedWebView *webView)
{ newPage(webView); }


void Browser::onDownloadStarted()
{
    downloads += 1;
    m_downloadLabel.show();
}


void Browser::onDownloadDone(QString &filepath, QString &mimetype)
{
    downloads -= 1;
    if (downloads < 0) /* shouldn't happen */
        downloads = 0;
    if (downloads == 0)
        m_downloadLabel.hide();
    m_processManager.open(filepath, mimetype);
}


void Browser::onMouseMove()
{ Q_EMIT mouseMoved(); }


void Browser::onEnd()
{
    stop();
    Q_EMIT end();
}


void Browser::nextTab()
{
    int idx = m_tabWidget.currentIndex() + 1;
    if (idx == m_tabWidget.count())
        idx = 0;
    m_tabWidget.setCurrentIndex(idx);
}


void Browser::prevTab()
{
    int idx = m_tabWidget.currentIndex() - 1;
    if (idx < 0)
        idx = m_tabWidget.count() - 1;
    m_tabWidget.setCurrentIndex(idx);
}


void Browser::selectAddressBar()
{
    NavBar *navbar = qobject_cast<Page *>(m_tabWidget.currentWidget())->navbar();
    navbar->locationEdit()->selectAll();
    navbar->locationEdit()->setFocus();
}


void Browser::selectAddressBarAndBeginSearch()
{
    NavBar *navbar = qobject_cast<Page *>(m_tabWidget.currentWidget())->navbar();
    navbar->locationEdit()->setText("?");
    navbar->locationEdit()->setCursorPosition(1);
    navbar->locationEdit()->setFocus();
}


void Browser::onUrlDenied(QString &url)
{
    m_statusBar.showMessage(QString("DENIED: %1").arg(url), 5000);
}


void Browser::addTime(int seconds)
{ remaining += seconds; }


void Browser::onTitleChanged(QString &title, Page *page)
{
    int idx = m_tabWidget.indexOf(page);
    m_tabWidget.setTabText(idx, title);
}

void Browser::mousePressEvent(QMouseEvent* event)
{
    if (event->button() == Qt::RightButton) {
        Q_EMIT rightclick();
    }
}

void Browser::hideToolbar()
{
    this->m_statusBar.hide();
    this->m_tabWidget.tabBar()->hide();
    this->m_cornerBtn.hide();
    qobject_cast<Page *>(this->m_tabWidget.currentWidget())->navbar()->hide();
}
