#include <unistd.h>
#include <sys/types.h>
#include <sys/utsname.h>

#include <QtWebKitWidgets>
#include <QLocale>
#include <QNetworkCookie>
#include <QVariant>
#include <QSqlQuery>
#include <QDateTime>
#include <QWebElementCollection>
#include <QWebFrame>
#include <QNetworkReply>
#include <QMouseEvent>
#include <QDir>

#include <QDebug>

#include "webpage.h"
#include "config.h"
#include "settings.h"


SqliteCookieJar::SqliteCookieJar(QObject *parent)
    : QNetworkCookieJar(parent)
{
    QSqlQuery query("SELECT name,value,host,path,expiry,isSecure,isHttpOnly FROM moz_cookies");
    while (query.next())
    {
        QNetworkCookie c;
        c.setName(query.value(0).toByteArray());
        c.setValue(query.value(1).toByteArray());
        c.setDomain(query.value(2).toByteArray());
        c.setPath(query.value(3).toByteArray());
        c.setExpirationDate(QDateTime::fromTime_t(query.value(4).toInt()));
        c.setSecure(query.value(5).toBool());
        c.setHttpOnly(query.value(6).toBool());
        m_cookies.append(c);
    }

    setAllCookies(m_cookies);
}


QString CustomQWebPage::userAgentForUrl(const QUrl &/*url*/) const
{
    Settings *settings = Settings::load();

    QString user_agent = settings->get("main/user_agent").toString();
    if (! user_agent.isEmpty())
        return user_agent;

    QLocale l;
    struct utsname sysinfo;
    uname(&sysinfo);

    user_agent = QString("Mozilla/5.0 (X11; %1; %2) AppleWebKit/%3 (KHTML, like Gecko) "
                         "Chrome/" CHROME_VERSION " Safari/%3 LinutopKiosk/%4")
                    .arg(sysinfo.sysname, l.name(), qWebKitVersion(), VERSION);

    return user_agent;
}


Webpage::Webpage(PlaylistItem *item, QWidget *parent)
    : QWebView(parent)
    , Runnable(item->time())
{
    m_uri = QUrl(item->uri());
    m_params = item->params();
    loaded = false;
    settings = Settings::load();
    m_increment = 1;
    m_scrollY = 0;
    m_scrollX = 0;

    setPage(&m_qwebpage);

//    QObject::connect(this, SIGNAL(loadFinished(bool)),
//                      static_cast<QWebView*>(this),
//                      SLOT(onLoadFinished(bool)));
    installEventFilter(this);

    m_jar = new SqliteCookieJar(this);
    page()->networkAccessManager()->setCookieJar(m_jar);

    connect(page()->networkAccessManager(), SIGNAL(finished(QNetworkReply*)), this, SLOT(errorCheck(QNetworkReply*)));

    page()->mainFrame()->setScrollBarPolicy(Qt::Vertical, Qt::ScrollBarAlwaysOff);
    page()->mainFrame()->setScrollBarPolicy(Qt::Horizontal, Qt::ScrollBarAlwaysOff);

    QObject::connect(
                page()->networkAccessManager(),
                SIGNAL(sslErrors(QNetworkReply*, const QList<QSslError> &)),
                this,
                SLOT(sslErrorHandler(QNetworkReply*, const QList<QSslError> &))
                );
    if(item->uri().endsWith(".txt"))
        displayHTML(item);

    QWebView::setContextMenuPolicy(Qt::CustomContextMenu);
    setZoomFactor(settings->get("main/zoomFactor").toDouble());

    bool as = false;

    Q_FOREACH(QString param, m_params)
    {
        if(param.contains("zoom:"))
        {
            setZoomFactor(param.split(":").at(1).toDouble());
        }
        if(param.contains("x:"))
        {
            m_scrollX = param.split(":").at(1).toInt();
            connect(page()->mainFrame(), SIGNAL(loadFinished(bool)), this, SLOT(scrollToX()));
        }
        if(param.contains("y:"))
        {
            m_scrollY = param.split(":").at(1).toInt();
            connect(page()->mainFrame(), SIGNAL(loadFinished(bool)), this, SLOT(scrollToY()));
        }
        if(param.contains("autoscroll"))
        {
            as = true;
        }
    }

    // Set black background as default on webpages in slideshow.
    QPalette pal = palette();
    pal.setBrush(QPalette::Base, Qt::black);
    page()->setPalette(pal);
    if(settings->get("slideshow/auto_scroll").toBool() || as)
    {
        connect(&m_startTimer, SIGNAL(timeout()), this, SLOT(startScroll()));
        connect(&m_scrollTimer, SIGNAL(timeout()), this, SLOT(autoScroll()));
        m_startTimer.start(5000);
    }
}

void Webpage::errorCheck(QNetworkReply *reply)
{
    disconnect(page()->networkAccessManager(), SIGNAL(finished(QNetworkReply*)), this, SLOT(errorCheck(QNetworkReply*)));
    if(reply->error() != QNetworkReply::NoError)
    {
        if(!m_params.contains("all_webpages"))
            Q_EMIT end();
        else
            setHtml("Network error: please make sure you are correctly connected to internet.");
    }
}

void Webpage::scrollToY()
{
    page()->mainFrame()->scroll(0, m_scrollY);
    disconnect(page()->mainFrame(), SIGNAL(loadFinished(bool)), this, SLOT(scrollToY()));
}

void Webpage::scrollToX()
{
    page()->mainFrame()->scroll(m_scrollX, 0);
    disconnect(page()->mainFrame(), SIGNAL(loadFinished(bool)), this, SLOT(scrollToX()));
}

void Webpage::startScroll()
{
    m_startTimer.stop();
    m_scrollTimer.start(50); //speed value of scrolling
    disconnect(&m_startTimer, SIGNAL(timeout()), this, SLOT(startScroll()));
}

void Webpage::autoScroll()
{
    int tmp = page()->mainFrame()->scrollPosition().ry();
    page()->mainFrame()->scroll(0,m_increment);
    int newPos = page()->mainFrame()->scrollPosition().ry();
    if(tmp == newPos)
        m_increment *= -1;
}

void Webpage::displayHTML(PlaylistItem *item)
{
    QString itemContent;
    QString tempContent;
    QFile itemFile(item->uri().mid(7));
    QFile baseTempFile("/usr/share/linutop-webkiosk/template.html");
    QFile templateFile(QDir::homePath() + "/.linutop-kiosk/current-template.html");

    if(itemFile.open(QIODevice::ReadOnly)
            && baseTempFile.open(QIODevice::ReadOnly)
            && templateFile.open(QIODevice::WriteOnly))
    {
        itemContent = itemFile.readAll();
        tempContent = baseTempFile.readAll();
        QString title;
        QString para;
        QStringList tmp;
        if(itemContent.startsWith("title=\""))
        {
            tmp = itemContent.split("\"");
            title = tmp.at(1);
            tempContent.replace("%title", title);
            tempContent.replace("display:none;", "");
            tmp = tmp.at(2).split("\n");
            for(int i=1; i<tmp.size()-1; i++)
                para.append(tmp.at(i)+"<br>");
        }
        else
        {
            tmp = itemContent.split("\n");
            for(int i=0; i<tmp.size()-1; i++)
                para.append(tmp.at(i)+"<br>");
        }

        tempContent.replace("%paragraph", para);
        templateFile.write(tempContent.toUtf8());
        itemFile.close(); baseTempFile.close(); templateFile.close();
        m_uri = QUrl(QDir::homePath() + "/.linutop-kiosk/current-template.html");
    }
}


void Webpage::sslErrorHandler(QNetworkReply *reply,
                              const QList<QSslError> &/*error_list*/)
{ reply->ignoreSslErrors(); }



bool Webpage::eventFilter(QObject *obj, QEvent *event)
{
    setMouseTracking(true);
    switch (event->type())
    {
        case QEvent::MouseMove:
        case QEvent::MouseButtonPress:
            this->mousePressEvent(dynamic_cast<QMouseEvent*>(event));
            return true;
        case QEvent::KeyPress:
            qobject_cast<QWidget *>(obj)->setCursor(Qt::BlankCursor);
            return true;
        default: break;
    }
    return false;
}


bool Webpage::startImpl(void)
{ load(m_uri); return true; }


bool Webpage::stopImpl(void)
{
    page()->action(QWebPage::Stop);
    QWebElementCollection elements = page()->mainFrame()->findAllElements("embed");
    Q_FOREACH (QWebElement e, elements)
        e.removeAllChildren();

    return true;
}


bool Webpage::onTimeoutImpl()
{ Q_EMIT end(); return true; }


void Webpage::pauseImpl(void) {}


void Webpage::resumeImpl(void) {}


void Webpage::mousePressEvent(QMouseEvent* event)
{
    switch(event->button()){
        case Qt::RightButton:
            Q_EMIT rightclick();
            break;
        default: break;
    }
}
