#include "settings.h"

#include <unistd.h>
#include <sys/types.h>
#include <pwd.h>

#include <fstream>

#include <QDebug>
#include <QFile>
#include <QTextStream>
#include <QProcess>
#include <QSettings>
#include <QUrl>
#include <QFont>
#include <QDateTime>
#include <QSqlDatabase>
#include <QSqlQuery>
#include <QDir>

#include "commandlineparser.h"
#include "version.h"


Settings *Settings::m_settings = NULL;

Settings *Settings::load(bool force, const CommandLineParser *cmdparse)
{
    if (force || !Settings::m_settings)
    {
        Settings::m_settings = new Settings(cmdparse);

        if (Settings::m_settings->m_version_changed ||
            Settings::m_settings->get("main/version").toString() !=
                LINUTOP_KIOSK_VERSION)
        {
            qDebug() << "Software has been updated. "
                        "Remove previous configuration.";

            if (QFile::exists(cmdparse->getPreferenceFilename()))
                QFile::remove(cmdparse->getPreferenceFilename());

            Settings::m_settings->m_map = QMap<QString, QVariant>();
            Settings::m_settings->populateDefaults();
            Settings::m_settings->writePreferences(
                cmdparse->getPreferenceFilename());
        }
    }


    return Settings::m_settings;
}

void Settings::populateDefaults()
{
    //////////
    // MAIN //
    //////////

    m_map["main/version"]                    = QVariant(LINUTOP_KIOSK_VERSION);
    m_map["main/admin_passwd"]               = QVariant("");
    m_map["main/after_web_session"]          = QVariant("");
    m_map["main/can_access_fs"]              = QVariant(false);
    m_map["main/can_end_session"]            = QVariant(true);
    m_map["main/can_print"]                  = QVariant(false);
    m_map["main/printer_name"]               = QVariant("");
    m_map["main/default_timer"]              = QVariant(900);
    m_map["main/default_url"]                = QVariant("http://www.linutop.com/");
    m_map["main/fullscreen"]                 = QVariant(true);
    m_map["main/hide_toolbar"]               = QVariant(false);
    m_map["main/fixedscreen"]                = QVariant(false);
    m_map["main/largeur_edit"]               = QVariant("");
    m_map["main/hauteur_edit"]               = QVariant("");
    m_map["main/list_mode"]                  = QVariant("none");
    m_map["main/list_use_regex"]             = QVariant(false);
    m_map["main/max_history"]                = QVariant(50);
    m_map["main/max_windows"]                = QVariant(0);
    // Can be "kiosk" or "display".
    m_map["main/mode"]                       = QVariant("display");
    m_map["main/new_tab_url"]                = m_map["main/default_url"];
    m_map["main/no_screensaver"]             = QVariant(true);
    m_map["main/post_session_uri"]           = QVariant(QUrl());
    m_map["main/post_session_uri_time"]      = QVariant(5);
    m_map["main/zoomFactor"]                 = QVariant(1.0),
    m_map["main/start_with_session"]         = QVariant(true);
    m_map["main/stop_session_after_seconds"] = QVariant(120);
    m_map["main/user_agent"]                 = QVariant("");
    m_map["main/with_clear"]                 = QVariant(true);
    m_map["main/with_timer"]                 = QVariant(false);
    m_map["main/external_browser"]           = QVariant(false);
    m_map["main/reset"]                      = QVariant(false);
#ifndef __arm__
    m_map["main/external_browser_cmd"]       = QVariant("firefox --new-window %u");
#else
    m_map["main/external_browser_cmd"]       = QVariant("chromium-browser -kiosk %u");
#endif

    ////////////
    // REMOTE //
    ////////////

    m_map["remote/direct"]        = QVariant(false);
    m_map["remote/enabled"]       = QVariant(false);
    m_map["remote/port"]          = QVariant(7575);
    m_map["remote/url"]           = QVariant("");
    m_map["remote/local_url"]     = QVariant("http://localhost:" + m_map["remote/port"].toString());
    m_map["remote/key"]           = QVariant("");
    m_map["remote/tv"]            = QVariant(true);
    m_map["remote/localplaylist"] = QVariant(true);
    m_map["remote/localfs"]       = QVariant(true);

    ///////////////
    // SLIDESHOW //
    ///////////////

    // Can be one of these:
    //   - localPlaylist
    //   - serverLinutopTV
    //   - serverLinutopTVPrivate
    //   - serverPlaylistPrivate
    m_map["slideshow/playlist_update_mode"]  = QVariant("localPlaylist");
    m_map["slideshow/default_time"]          = QVariant(5);
    m_map["slideshow/resize"]                = QVariant(true);
    m_map["slideshow/stretch"]               = QVariant(false);
    m_map["slideshow/auto_scroll"]           = QVariant(false);
    m_map["slideshow/do_local_update"]       = QVariant(false);
    m_map["slideshow/hide_pointer"]          = QVariant(true);
    m_map["slideshow/mute"]                  = QVariant(false);
    m_map["slideshow/refresh_after_minutes"] = QVariant(30);
    m_map["slideshow/save_modifications"]    = QVariant(false);
    m_map["slideshow/stop_on_event"]         = QVariant(true);
    m_map["slideshow/show_address"]          = QVariant(true);
    m_map["slideshow/clean_after"]           = QVariant(false);
    m_map["slideshow/usb_auto"]              = QVariant(false);
    m_map["slideshow/shuffle"]               = QVariant(false);
    m_map["slideshow/use_remote"]            = QVariant(false);
    m_map["slideshow/invoice"]               = QVariant("");
    m_map["slideshow/email"]                 = QVariant("");
    m_map["slideshow/url"]                   = QVariant("");

#ifdef __arm__
    m_map["slideshow/external_player"]       = QVariant(true);
#else
    m_map["slideshow/external_player"]       = QVariant(false);
#endif
    m_map["slideshow/external_browser"]      = QVariant(false);

    /////////////
    // WEBPAGE //
    /////////////

    m_map["webpage/default_font"] =
            QVariant(QFont("Droid Sans",14,-1, false));

    /////////////
    // NETWORK //
    /////////////

    m_map["network/proxy_user"]   = QVariant("");
    m_map["network/proxy_passwd"] = QVariant("");
    m_map["network/proxy_type"]   = QVariant(0);
    m_map["network/proxy_uri"]    = QVariant("");

}

Settings::Settings(CommandLineParser const* cmdparse)
    : QObject()
    , m_cmdparse(cmdparse)
    , m_version_changed(true)
{
    populateDefaults();
    readAutostarted();

    QString conf_filename = cmdparse->getPreferenceFilename();
    QSettings tmp(conf_filename, QSettings::IniFormat);
    QStringList allkeys = tmp.allKeys();
    qDebug() << "Loading settings from:" << conf_filename;
    Q_FOREACH (const QString &k, allkeys)
    {
        qDebug() << k << "=" << tmp.value(k);
        m_map[k] = tmp.value(k);
    }

    if (!allkeys.contains("main/version"))
        m_version_changed = true;
    else
        m_version_changed = false;
}

void Settings::save(bool with_autostart,
                    const QString &settings,
                    const QString &playlist,
                    const QString &restrictions,
                    const QString &cookies)
{
    writePreferences(settings);
    writePlaylist(playlist);
    writeRestrictions(restrictions);
    writeCookies(cookies);

    enableScreenSaver(!m_map["main/no_screensaver"].toBool());

    if (with_autostart) writeAutostarted();
}

void Settings::save(bool with_autostart, const QString &basePath)
{
    save(with_autostart,
         CommandLineParser::defaultPreferenceFilename(basePath),
         CommandLineParser::defaultPlaylistFilename(basePath),
         CommandLineParser::defaultWhiteListFilename(basePath),
         CommandLineParser::defaultCookieFilename(basePath));
}

void Settings::save(bool with_autostart)
{ save(with_autostart, CommandLineParser::defaultBasePath()); }

bool Settings::writePreferences(QString const& where)
{
    QString where2 = where.isEmpty()
            ? m_cmdparse->getPreferenceFilename()
            : where;

    qDebug() << "Saving preference file at:" << where2;

    QSettings tmp(where2, QSettings::IniFormat);
    if (!tmp.isWritable())
    {
        qDebug() << "Unable to write preference file.";
        return false;
    }

    Q_FOREACH (QString key, m_map.keys())
    {
        QStringList spl = key.split("/");
        if (spl.at(1) == "size") /* FIXME: where does this come from? */
            continue;

        tmp.beginWriteArray(spl.at(0));
        qDebug() << spl.at(0) << spl.at(1) << m_map[key];
        tmp.setValue(spl.at(1), m_map[key]);
        tmp.endArray();
    }

    tmp.sync();

    return true;
}

bool Settings::writeCookies(QString const& where)
{
    QString where2 = where.isEmpty() ? cookiesFilename() : where;

    qDebug() << "Saving cookie file at:" << where2;

    QSqlDatabase db = QSqlDatabase::addDatabase("QSQLITE", "cookies");
    db.setDatabaseName(where2);
    db.open();
    QSqlQuery q(db);

    db.transaction();
    q.exec("DROP TABLE IF EXISTS moz_cookies");
    q.exec("CREATE TABLE moz_cookies (id INTEGER PRIMARY KEY, "
                "name TEXT, value TEXT, host TEXT, path TEXT, "
                "expiry INTEGER, lastAccessed INTEGER, isSecure INTEGER,"
                "isHttpOnly INTEGER)");
    Q_FOREACH (const QNetworkCookie cookie, cookiesList)
    {
        int isSecure = 0;
        if (cookie.isSecure()) isSecure = 1;
        int isHttpOnly = 0;
        if (cookie.isHttpOnly()) isHttpOnly = 1;

        QString sql("INSERT INTO moz_cookies "
                    "(name, value, host, path,"
                    "expiry, lastAccessed, isSecure, isHttpOnly) "
                    "VALUES ('%1', '%2', '%3', '%4', '%5', '', '%6', '%7')");
        QString result = sql.arg(QString(cookie.name()))
                .arg(QString(cookie.value()))
                .arg(QString(cookie.domain()))
                .arg(QString(cookie.path()))
                .arg(cookie.expirationDate().toTime_t())
                .arg(isSecure)
                .arg(isHttpOnly);
        q.exec(result);
    }
    db.commit();
    db.close();

    return true;
}

bool Settings::writePlaylist(QString const& where)
{
    QString where2 = where.isEmpty() ? playlistFilename() : where;

    qDebug() << "Saving preference file at:" << where2;

    QFile f(where2);
    if (!f.open(QIODevice::WriteOnly)) return false;

    QTextStream out(&f);
    out << "#Linutop playlist\n";
    Q_FOREACH (QString line, playlistItems) out << line << "\n";

    f.close();
    return true;
}

bool Settings::writeRestrictions(QString const& where)
{
    QString where2 = where.isEmpty() ? whileListFilename() : where;

    qDebug() << "Saving restriction file at:" << where2;

    QFile f(where2);
    if (!f.open(QIODevice::WriteOnly)) return false;

    QTextStream out(&f);
    Q_FOREACH (QString line, restrictions) out << line << "\n";

    f.close();
    return true;
}

void Settings::readAutostarted()
{
    m_autostarted = false;
    QProcess p;
    QStringList args;
    args << "--selected";
    p.start("appstarter", args);
    p.waitForFinished();
    if (p.readAllStandardOutput().startsWith("webkiosk"))
        m_autostarted = true;
}

void Settings::writeAutostarted()
{
#ifdef __arm__
QString cfile = "/home/pi/.config/appstarterrc";
#else
QString cfile = "/home/linutop/.config/appstarterrc";
#endif

    QString what = m_autostarted ? "webkiosk" : "none";
    QString cmd = m_autostarted ? "linutop-kiosk" : "true";
    QStringList args;
    args << "--select" << what << "--command" << cmd << "-c" << cfile;
    QProcess p;
    p.start("appstarter", args);
    p.waitForFinished();
    QByteArray uid = qgetenv("SUDO_UID");
    if (! uid.isEmpty())
        if(chown(cfile.toUtf8().data(), uid.toInt(), -1))
            qDebug() << "chown done.";
}

void Settings::enableScreenSaver(bool enable)
{
    QStringList args;
    if(enable)
    {
        QProcess p;
        p.start("xset",args<<"s"<<"blank");
        p.waitForFinished();
        p.start("xset",args<<"s"<<"on");
        p.waitForFinished();
        p.start("xset",args<<"+dpms");
        p.waitForFinished();
    }
    else{

        QProcess p;
        p.start("xset",args<<"s"<<"noblank");
        p.waitForFinished();
        p.start("xset",args<<"s"<<"off");
        p.waitForFinished();
        p.start("xset",args<<"-dpms");
        p.waitForFinished();
    }
}

QString Settings::preferencesFilename(void) const
{ return m_cmdparse->getPreferenceFilename(); }

QString Settings::playlistFilename(void) const
{ return m_cmdparse->getPlaylistFilename(); }

QString Settings::whileListFilename(void) const
{ return m_cmdparse->getWhiteListFilename(); }

QString Settings::cookiesFilename(void) const
{ return m_cmdparse->getCookieFilename(); }

QString Settings::hookDirname(void) const
{ return m_cmdparse->getHookDirname(); }

QVariant Settings::get(const QString &key) const
{ return m_map[key]; }

void Settings::dump()
{
    QMapIterator<QString, QVariant> i(m_map);
    while (i.hasNext())
    {
        i.next();
        qDebug() << i.key() << ": " << i.value();
    }
    qDebug() << "hookdir: " << hookDirname();
}

void Settings::set(const QString &key, QString value)
{ m_map[key] = QVariant(value); }

void Settings::set(const QString &key, bool value)
{ m_map[key] = value; }

void Settings::set(const QString &key, int value)
{ m_map[key] = value; }

void Settings::set(const QString &key, double value)
{ m_map[key] = value; }
