# -*- coding: utf-8 -*-
import os.path
import json
import requests

import ConfigParser
from bottle_logging import logger


class Settings:
    def __init__(self):
        self.config = ConfigParser.ConfigParser()
        self.conffile = "/etc/xdg/webkiosk/preferences"
        if os.path.isfile("/tmp/linutop-kiosk-test/preferences"):
            self.conffile = "/tmp/linutop-kiosk-test/preferences"

    def _read_prefs(self):
        self.config.read(self.conffile)

    def _write_prefs(self):
        with open(self.conffile, 'w') as fd:
            self.config.write(fd)

    def testServerConnection(self, url, invoice, email):
        headers = {"Content-type":
                   "application/x-www-form-urlencoded; charset=UTF-8",
                   "Accept": "application/json"}
        url += "/webapp/api/user/" + invoice
        data = {
            "model": json.dumps({
                "invoice": invoice,
                "email": email
            })
        }

        try:
            response = requests.post(url, data=data, headers=headers)
            return response.status_code == requests.codes.ok
        except requests.ConnectionError as e:
            logger.error("Fail to connect: %s" % str(e))
            return False

        return True

    def getServerType(self):
        """Find what to display as server.

        Find if the server to display is a:
        - serverLinutopTVPrivate
        - serverPlaylistPrivate
        - serverLinutopTV

        This function will be used to display the selectable channels in the remote interface.

        Returns:
            string:serverType, string:serverUrl, string:invoice, string:email

            serverType is one of the following:
            - serverLinutopTVPrivate
            - serverPlaylistPrivate
            - serverLinutopTV

            serverUrl is the url to get connection with the remote server.

            invoice is the id to connect.

            email is the password to connect.
        """

        self._read_prefs()

        try:
            invoice = self.config.get("slideshow", "invoice")
        except ConfigParser.NoOptionError:
            invoice = ""
        try:
            email = self.config.get("slideshow", "email")
        except ConfigParser.NoOptionError:
            email = ""
        try:
            mode = self.config.get("slideshow", "playlist_update_mode")
        except ConfigParser.NoOptionError:
            mode = "localPlaylist"

        # Try Linutop TV private server.
        try:
            linutopTVPrivateUrl = self.config.get("slideshow", "url")
            if linutopTVPrivateUrl and self.testServerConnection(linutopTVPrivateUrl, invoice, email):
                return "serverLinutopTVPrivate", linutopTVPrivateUrl, invoice, email
            else:
                logger.info("Authentication at linutopPrivateServer: %s failed." % linutopTVPrivateUrl)
        except ConfigParser.NoOptionError:
            pass

        # Try Linutop.tv
        serverLinutopTVUrl = "http://www.linutop.tv"
        if self.testServerConnection(serverLinutopTVUrl, invoice, email):
            return "serverLinutopTV", serverLinutopTVUrl, invoice, email
        else:
            logger.info("Authentication at serverLinutopTV: %s failed." % serverLinutopTVUrl)

        return "serverLinutopTV", serverLinutopTVUrl, "", ""

    def setChannel(self, serverType, url):
        self.config.set("slideshow", "pages", url)
        self.config.set("slideshow", "use_remote", 'true')
        self.config.set("slideshow", "playlist_update_mode", serverType)
        self._write_prefs()

    def getRemote(self):
        self._read_prefs()
        try:
            return self.config.get("slideshow", "playlist_update_mode") != 'localPlaylist'
        except ConfigParser.NoOptionError:
            return False

    def setLocal(self):
        self.config.set("slideshow", "playlist_update_mode", 'localPlaylist')
        self._write_prefs()

    def getDirect(self):
        self._read_prefs()
        try:
            return self.config.get("remote", "direct") == 'true'
        except ConfigParser.NoOptionError:
            return False

    def getKey(self):
        self._read_prefs()
        try:
            return self.config.get("remote", "key")
        except ConfigParser.NoOptionError:
            return ""

    def isLocalPlaylistFile(self):
        return os.path.isfile(os.path.join("/etc", "xdg", "webkiosk", "testplaylist.txt"))

    def getLocalPlaylistFile(self):
        if self.isLocalPlaylistFile():
            playlists = []
            ids = []
            with open(os.path.join("/etc", "xdg", "webkiosk", "testplaylist.txt")) as f:
                for line in f.readlines():
                    if line == "\n":
                        continue
                    line = tuple(line.strip("\n\r").split(' '))
                    if len(line) < 2:
                        continue
                    if line[1].find("http") == 0:
                        playlists.append(line)
                    else:
                        ids.append(line)
                ids.append(playlists)
                return ids
        return []

    def isTv(self):
        self._read_prefs()
        try:
            return self.config.get("remote", "tv") == "true"
        except ConfigParser.NoOptionError:
            return True

    def isLocalPlaylist(self):
        self._read_prefs()
        try:
            return self.config.get("remote", "localplaylist") == "true"
        except ConfigParser.NoOptionError:
            return True

    def isLocalFs(self):
        self._read_prefs()
        try:
            return self.config.get("remote", "localfs") == "true"
        except ConfigParser.NoOptionError:
            return True
